# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/ping.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import pydantic

from pyknic.lib.bellboy.app import BellBoyCommandHandler, register_bellboy_command, LobbyClient
from pyknic.lib.bellboy.models import OptionalBellBoyCommandModel
from pyknic.lib.fastapi.lobby import LobbyCommandHandler, register_lobby_command
from pyknic.lib.fastapi.models.base import NullableModel
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyStrFeedbackResult, LobbyCommandRequest

from pyknic.version import __version__


@register_lobby_command()
class LobbyPingCommand(LobbyCommandHandler):
    """This is a server side API-"command" that takes no arguments and returns internal version
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.LobbyCommandHandler.command_name` method implementation
        """
        return "ping"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.LobbyCommandHandler.command_model` method implementation
        """
        return NullableModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.LobbyCommandHandler.exec` method implementation
        """
        return LobbyStrFeedbackResult(
            str_result=f"pong... Version is {__version__}"
        )


@register_bellboy_command()
class BellBoyPingCommand(BellBoyCommandHandler):
    """This is a CLI "command" that returns client's version (if there wasn't any argument) or server's version (if
    URL and secret backend were set)
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.BellBoyCommandHandler.command_name` method implementation
        """
        return LobbyPingCommand.command_name()

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.BellBoyCommandHandler.command_model` method implementation
        """
        return OptionalBellBoyCommandModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.BellBoyCommandHandler.exec` method implementation
        """
        assert(isinstance(self._args, OptionalBellBoyCommandModel))

        if self._args.server is None:
            result = await LobbyPingCommand.prepare_command(NullableModel()).exec()
            result.str_result = f"Client response: {result.str_result}"  # type: ignore[union-attr]
            return result

        auth_data = self.auth_data(self._args.server.secret_backend, self._args.server.lobby_url)

        client = LobbyClient(self._args.server.lobby_url, auth_data.server_fingerprint, auth_data.token)
        result = await client.command_request(
            LobbyCommandRequest(
                name=LobbyPingCommand.command_name(),
                args=NullableModel().model_dump()
            )
        )

        result.str_result = f"Server response: {result.str_result}"  # type: ignore[union-attr]
        return result



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/collection.py
#
# Copyright (C) 2017-2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.capability import iscapable
from pyknic.lib.registry import APIRegistry
from pyknic.lib.io.clients.proto import IOClientProto
from pyknic.lib.verify import verify_value
from pyknic.lib.uri import URI


__default_io_clients_registry__ = APIRegistry()


class IOVirtualClient(IOClientProto):

    @verify_value(uri=lambda x: x.scheme is not None)
    def __init__(self, uri: URI, registry: typing.Optional[APIRegistry] = None) -> None:
        IOClientProto.__init__(self)

        self.__uri = uri

        if registry is None:
            registry = __default_io_clients_registry__

        self.__client = registry.get(uri.scheme).create_client(uri)
        self.__init_capability('connect')
        self.__init_capability('disconnect')
        self.__init_capability('change_directory')
        self.__init_capability('list_directory')
        self.__init_capability('make_directory')
        self.__init_capability('remove_directory')
        self.__init_capability('upload_file')
        self.__init_capability('remove_file')
        self.__init_capability('receive_file')
        self.__init_capability('file_size')

    def __init_capability(self, method_name: str) -> None:
        cap = getattr(IOClientProto, method_name)
        if iscapable(self.__client, cap):
            self.append_capability(cap, getattr(self.__client, method_name))

    @classmethod
    def create_client(cls, uri: URI) -> 'IOClientProto':
        return cls(uri, registry=__default_io_clients_registry__)

    def uri(self) -> URI:
        return self.__uri



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/base.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code

import gettext
import fastapi
import typing

from abc import abstractmethod

from pyknic.lib.config import Config
from pyknic.lib.gettext import GetTextWrapper
from pyknic.lib.fastapi.apps_registry import FastAPIAppProto
from pyknic.lib.fastapi.models.tg_bot_types import User, Message, ReplyParameters, Update
from pyknic.lib.fastapi.models.tg_bot_methods import MethodSendMessage, MethodAnswerCallbackQuery
from pyknic.lib.fastapi.models.base import NullableModel


# noinspection PyAbstractClass
class BaseFastAPIApp(FastAPIAppProto):
    """ Base implementation for FastAPI apps
    """

    def __init__(self, config: Config, translations: GetTextWrapper):
        """ Create an app

        :param config: The pyknic config
        :param translations: application translation
        """
        FastAPIAppProto.__init__(self)
        self.__config = config
        self.__translations = translations

    def config(self) -> Config:
        """ Return global config
        """
        return self.__config

    def lang(self, lang_name: str | None = None) -> gettext.GNUTranslations | gettext.NullTranslations:
        """ Return translation for the specified language

        :param lang_name: a language to return (a default one will be used if this parameter is not set)
        """
        return self.__translations(lang_name)


TgBotResponseType: typing.TypeAlias = typing.Union[MethodSendMessage, MethodAnswerCallbackQuery, NullableModel]


class TgBotBaseFastAPIApp(BaseFastAPIApp):
    """ Base implementation for telegram bot
    """

    @classmethod
    def create_app(cls, fastapi_app: fastapi.FastAPI, config: Config, translations: GetTextWrapper) -> typing.Any:
        """ The :meth:`.FastAPIAppProto.create_app` method implementation
        """
        app = cls(config, translations)

        # :note: response_model_exclude_unset=True excludes defaults
        # :note: null values must be excluded
        # :note: non-null defaults should be returned
        fastapi_app.post(
            cls.bot_path(config),
            status_code=200,
            response_model=TgBotResponseType,
            response_model_exclude_none=True
        )(app.tgbot_webhook_request)

        return app

    @classmethod
    @abstractmethod
    def bot_path(cls, config: Config) -> str:
        """ Return an HTTP path of webhook that this bot uses
        """
        raise NotImplementedError('This method is abstract')

    async def tgbot_webhook_request(self, request: fastapi.Request) -> TgBotResponseType:
        """ Process request from a Telegram's server. This method just parses a input and all the work is done inside
        the :meth:`.TgBotBaseFastAPIApp.process_request` method
        """
        data = await request.json()
        tg_update = Update.model_validate(data)
        return await self.process_request(tg_update)

    async def process_request(self, tg_update: Update) -> TgBotResponseType:
        """ Try to process a request. Request is processed in the following order:
          - first of all if the request has a callback_query then
          the :meth:`TgBotBaseFastAPIApp.callback_query` method result is returned
          - then if there is a message that has text and this text starts with the "/" then this message will be
          treated as a command and the :meth:`TgBotBaseFastAPIApp.process_command` method result is checked
          (if there is one, then return it)
          - then if there is a message that has text then try to process it with
          the :meth:`TgBotBaseFastAPIApp.process_message` method
          - at the end a "null" ("{}" -- empty dict) result is returned
        """
        if tg_update.callback_query is not None:
            return await self.callback_query(tg_update)  # TODO: everything else to async!

        if tg_update.message is not None and tg_update.message.text is not None and tg_update.message.from_ is not None:
            msg = tg_update.message.text.strip().lower()

            if msg.startswith('/') and len(msg) > 1:
                result = await self.process_command(msg, tg_update)
                if result is not None:
                    return result

            result = await self.process_message(tg_update)
            if result is not None:
                return result

        return NullableModel()

    async def callback_query(self,  tg_update: Update) -> MethodAnswerCallbackQuery:
        """ A request treated as a callback_query -- return a default value
        """
        assert(tg_update.callback_query is not None)
        return MethodAnswerCallbackQuery(callback_query_id=str(tg_update.callback_query.id_))

    async def process_command(self, command: str, tg_update: Update) -> TgBotResponseType | None:
        """ A request is a command -- return a result if command is valid and return None otherwise
        """
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)
        return None

    async def process_message(self, tg_update: Update) -> TgBotResponseType | None:
        """ A request is just a text -- try to process it
        """
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)
        return None

    def user_lang(self, user: User | None = None) -> gettext.GNUTranslations | gettext.NullTranslations:
        """ Return a translation that is based on a users language
        """
        return self.lang(user.language_code if user is not None else None)

    def reply(self, tg_obj: Message | Update, text: str, **kwargs: typing.Any) -> MethodSendMessage:
        """ Shortcut for a response that replies to user's message

        :param tg_obj: original user request
        :param text: text to reply
        :param kwargs: extra parameters for the MethodSendMessage object (besides "chat_id", "text"
        and "reply_parameters" fields)
        """

        tg_msg = tg_obj if isinstance(tg_obj, Message) else tg_obj.message
        assert(tg_msg is not None)

        return MethodSendMessage(
            chat_id=tg_msg.chat.id_,
            text=text,
            reply_parameters=ReplyParameters(message_id=tg_msg.message_id),
            **kwargs
        )



# -*- coding: utf-8 -*-
# pyknic/lib/io/aio_wrapper.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import functools
import inspect
import time
import typing

from pyknic.lib.tasks.proto import TaskProto
from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.thread_executor import ThreadExecutor
from pyknic.lib.tasks.threaded_task import ThreadedTask
from pyknic.lib.signals.extra import AsyncWatchDog
from pyknic.lib.verify import verify_value
from pyknic.lib.io import __default_block_size__, IOGenerator, IOAsyncGenerator, IOProcessor, IOAsyncProcessor
from pyknic.lib.io import IOProducer


class AsyncWrapper:
    """This class wraps thread-running and async awaiting routine into a "single" call."""

    # TODO: make it to work with TaskProto.stop and TaskProto.terminate

    def __init__(
        self,
        task: TaskProto,
        loop: asyncio.AbstractEventLoop,
        thread_executor: typing.Optional[ThreadExecutor] = None
    ) -> None:
        """Create a new wrapper

        :param task: the task to start
        :param loop: loop inside which call will be made
        :param thread_executor: thread executor to execute a task with (if None then a new
        thread will be created)
        """
        self.__task = task
        self.__watchdog = AsyncWatchDog(loop, self.__task, TaskProto.task_completed)
        self.__executor = thread_executor

    async def __call__(self) -> typing.Any:
        """Execute a task in asynchronous way."""

        if self.__executor is not None:
            task_executor = self.__executor.start_async(self.__task)
        else:
            threaded_task = ThreadedTask(self.__task)
            task_executor = threaded_task.start_async()

        watchdog_future = await self.__watchdog.watchdog_future()
        await asyncio.gather(task_executor, watchdog_future)
        watchdog_result = watchdog_future.result()

        if watchdog_result.value.exception:
            raise watchdog_result.value.exception

        return watchdog_result.value.result

    @classmethod
    async def create(
        cls,
        fn: typing.Callable[[], typing.Any],
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
        thread_executor: typing.Optional[ThreadExecutor] = None
    ) -> "AsyncWrapper":
        """Create a new wrapper

        :param fn: the function to start (to wrap)
        :param loop: same as loop in the `.meth:AsyncWrapper.__init__` method
        :param thread_executor: same as thread_executor in the `.meth:AsyncWrapper.__init__` method
        """

        return cls(PlainTask(fn), loop if loop else asyncio.get_event_loop(), thread_executor)


class IOThrottler:
    """ This class helps to pause for limiting IO requests
    """

    @verify_value(throttling=lambda x: x is None or x > 0)
    def __init__(self, throttling: typing.Optional[typing.Union[int, float]] = None):
        """ Create a new throttler

        :param throttling: number of bytes per seconds this throttler should aim to
        """
        self.__throttling = throttling
        self.__start_point: typing.Optional[float] = None
        self.__processed_bytes = 0

    def start(self) -> None:
        """ Start timer for this throttler. May be started only once!
        """
        if self.__start_point is not None:
            raise RuntimeError('This throttler has already started.')

        if self.__throttling is not None:
            self.__start_point = time.monotonic()

        self.__processed_bytes = 0

    def __iadd__(self, other: int) -> "IOThrottler":
        """Increase number of already processed bytes

        :param other: number of bytes
        """
        self.__processed_bytes += other
        return self

    def pause(self) -> typing.Union[int, float]:
        """Return required pause this throttler needs
        """
        if self.__throttling is None:
            return 0

        if self.__start_point is None:
            raise RuntimeError('This throttler was not started.')

        elapsed = time.monotonic() - self.__start_point
        should_be = float(self.__processed_bytes) / self.__throttling
        return (should_be - elapsed) if should_be > elapsed else 0

    @verify_value(read_size=lambda x: x is None or x > 0, block_size=lambda x: x is None or x > 0)
    def reader(
        self,
        io_obj: typing.IO[bytes],
        *,
        read_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None
    ) -> IOGenerator:
        """A basic reader implementation that reads file-object in blocking mode, and yields chunks of data. A pause
        should be made after every chunk in order to respect throttling settings. The pause duration may be found
        with the :meth:`.IOThrottler.pause` method.

        :note: This or :meth:`.IOThrottler.writer`, :meth:`.IOThrottler.copier`, :meth:`.IOThrottler.start` methods
        may be called only once.

        :param io_obj: IO-object to read
        :param read_size: total number of bytes to read
        :param block_size: number of bytes to read on each blocking read request
        """

        if block_size is None:
            block_size = __default_block_size__

        if read_size is None:
            read_size = -1

        self.start()

        next_block = io_obj.read(block_size if read_size < 0 else min(block_size, read_size))

        while next_block and (not not read_size):
            block_len = len(next_block)
            self.__processed_bytes += block_len
            read_size -= block_len
            yield next_block

            next_block = io_obj.read(block_size if read_size < 0 else min(block_size, read_size))

        if read_size > 0:
            raise ValueError(f'Source object does not have enough data. "{read_size}" bytes are missing')

    @staticmethod
    def sync_reader(
        io_obj: typing.IO[bytes],
        *,
        read_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOGenerator:
        """A blocking variant of :meth:`IOThrottler.reader` method. It sleeps in blocking mode in order to
        respect throttling settings. It is useful for running in a dedicated thread
        """

        throttler = IOThrottler(throttling)
        for block in throttler.reader(io_obj, read_size=read_size, block_size=block_size):
            yield block
            time.sleep(throttler.pause())

    @staticmethod
    async def async_reader(
        io_obj: typing.IO[bytes],
        *,
        read_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOAsyncGenerator:
        """An asyncio variant of :meth:`IOThrottler.reader` method. It sleeps in asyncio mode in order to
        respect throttling settings.
        """

        throttler = IOThrottler(throttling)
        for block in throttler.reader(io_obj, read_size=read_size, block_size=block_size):
            yield block
            await asyncio.sleep(throttler.pause())

    @verify_value(write_size=lambda x: x is None or x > 0, block_size=lambda x: x is None or x > 0)
    def writer(
        self,
        source: IOProducer,
        io_obj: typing.IO[bytes],
        *,
        write_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None
    ) -> IOGenerator:
        """A basic writer implementation that writes file-object in blocking mode, and yields chunks of data that
        has been written. A pause should be made after every chunk in order to respect throttling settings.
        The pause duration may be found with the :meth:`.IOThrottler.pause` method.

        :note: This or :meth:`.IOThrottler.writer`, :meth:`.IOThrottler.copier`, :meth:`.IOThrottler.start` methods
        may be called only once.

        :param source: generator that yields data to write
        :param io_obj: IO-object to write to
        :param write_size: total number of bytes to write
        :param block_size: a maximum number of bytes to write on each blocking write request
        """

        def data_splitter(generated_data: bytes) -> IOGenerator:
            nonlocal block_size

            if block_size is None:
                block_size = __default_block_size__

            start_pos = 0
            data_left = len(generated_data)
            while data_left > 0:
                yield generated_data[start_pos:(start_pos + block_size)]
                start_pos += block_size
                data_left -= block_size

        if write_size is None:
            write_size = -1

        self.start()

        for data in source:
            for next_block in data_splitter(data):
                next_block_len = len(next_block)

                if 0 < write_size < next_block_len:
                    next_block = next_block[:write_size]
                    next_block_len = write_size

                io_obj.write(next_block)
                yield next_block

                self.__processed_bytes += next_block_len
                write_size -= next_block_len

            if not write_size:
                return

        if write_size > 0:
            raise ValueError(f'Source object does not have enough data. "{write_size}" bytes are missing')

    @staticmethod
    def sync_writer(
        source: IOProducer,
        io_obj: typing.IO[bytes],
        *,
        write_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOGenerator:
        """A blocking variant of :meth:`IOThrottler.writer` method. It sleeps in blocking mode in order to
        respect throttling settings. It is useful for running in a dedicated thread
        """

        throttler = IOThrottler(throttling)
        for written_chunk in throttler.writer(source, io_obj, write_size=write_size, block_size=block_size):
            yield written_chunk
            time.sleep(throttler.pause())

    @staticmethod
    async def async_writer(
        source: IOProducer,
        io_obj: typing.IO[bytes],
        *,
        write_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOAsyncGenerator:
        """An asyncio variant of :meth:`IOThrottler.__writer` method. It sleeps in asyncio mode in order to
        respect throttling settings
        """

        throttler = IOThrottler(throttling)
        for written_chunk in throttler.writer(source, io_obj, write_size=write_size, block_size=block_size):
            yield written_chunk
            await asyncio.sleep(throttler.pause())

    @verify_value(copy_size=lambda x: x is None or x > 0, block_size=lambda x: x is None or x > 0)
    def copier(
        self,
        source_io_obj: typing.IO[bytes],
        destination_io_obj: typing.IO[bytes],
        *,
        copy_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None
    ) -> IOGenerator:
        """A basic data copier implementation that copies data from one IO-object to other one. A pause should be made
        after every chunk in order to respect throttling settings. The pause duration may be found with
        the :meth:`.IOThrottler.pause` method. This method yields successfully copied chunks.

        :note: This or :meth:`.IOThrottler.writer`, :meth:`.IOThrottler.copier`, :meth:`.IOThrottler.start` methods
        may be called only once.

        :param source_io_obj: IO-object to read
        :param destination_io_obj: IO-object to write to
        :param copy_size: number of bytes to copy
        :param block_size: (optional) number of bytes to read/write on each blocking request
        """

        reader_throttler = IOThrottler()
        self.start()

        for block in reader_throttler.reader(source_io_obj, read_size=copy_size, block_size=block_size):
            destination_io_obj.write(block)
            self.__processed_bytes += len(block)

            yield block

    @staticmethod
    def sync_copier(
        source_io_obj: typing.IO[bytes],
        destination_io_obj: typing.IO[bytes],
        *,
        copy_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOGenerator:
        """A blocking variant of :meth:`IOThrottler.copier` method. It sleeps in blocking mode in order to
        respect throttling settings. It is useful for running in a dedicated thread
        """

        throttler = IOThrottler(throttling)

        for block in throttler.copier(source_io_obj, destination_io_obj, copy_size=copy_size, block_size=block_size):
            yield block
            time.sleep(throttler.pause())

    @staticmethod
    async def async_copier(
        source_io_obj: typing.IO[bytes],
        destination_io_obj: typing.IO[bytes],
        *,
        copy_size: typing.Optional[int] = None,
        block_size: typing.Optional[int] = None,
        throttling: typing.Optional[int] = None
    ) -> IOAsyncGenerator:
        """An asyncio variant of :meth:`IOThrottler.copier` method. It sleeps in asyncio mode in order to
        respect throttling settings.
        """

        throttler = IOThrottler(throttling)

        for block in throttler.copier(source_io_obj, destination_io_obj, copy_size=copy_size, block_size=block_size):
            yield block
            await asyncio.sleep(throttler.pause())


def cg(source: IOGenerator) -> int:
    """Run over an IO generator and complete it.

    :param source: a generator that yields chunks of data.

    :returns: a total number of bytes
    """
    return sum((len(x) for x in source))


async def as_ag(source: IOGenerator) -> IOAsyncGenerator:
    """Convert general generator to an async one

    :param source: generator to convert
    """
    for i in source:
        yield i


async def cag(source: typing.Union[IOGenerator, IOAsyncGenerator]) -> int:
    """Run over an "IO processor" and complete it.

    :param source: "IO processor" that yields chunks of data.

    :returns: a total number of bytes
    """
    result = 0

    async for data in (source if inspect.isasyncgen(source) else as_ag(source)):  # type: ignore[arg-type]
        result += len(data)
    return result


def chain_sync_processor(source: IOGenerator, *processors: IOProcessor) -> IOGenerator:
    """This function runs chunks from source (general generator) through the processors and yields result.

    :param source: ordinary generator that yields chunks of data to process.
    :param processors: processors to run
    """

    def dummy_processor(s: IOGenerator) -> IOGenerator:
        # this is just to make sure that there is one processor at least
        for i in s:
            yield i

    chain = functools.partial(dummy_processor, source)
    for processor in processors:
        chain = functools.partial(processor, chain())

    for chunk in chain():
        yield chunk


async def chain_async_processor(
    source: typing.Union[IOGenerator, IOAsyncGenerator], *processors: IOAsyncProcessor
) -> IOAsyncGenerator:
    """This function runs chunks from source (general generator or async generator) through the processors
    and yields result.

    :param source: generator (ordinary or async one) that yields chunks of data to process.
    :param processors: processors to run
    """

    async def dummy_processor(s: typing.Union[IOGenerator, IOAsyncGenerator]) -> IOAsyncGenerator:
        # this is just to make sure that there is one processor at least
        async for i in (s if inspect.isasyncgen(source) else as_ag(s)):  # type: ignore[arg-type, union-attr]
            yield i

    chain = functools.partial(dummy_processor, source)
    for processor in processors:
        chain = functools.partial(processor, chain())

    async for chunk in chain():
        yield chunk



# -*- coding: utf-8 -*-

import gc
import functools
import pytest
import threading
import typing

from pyknic.lib.thread import acquire_lock, critical_section_dynamic_lock, critical_section_lock, CriticalResource
from pyknic.lib.thread import CriticalSectionError


def test_acquire_lock() -> None:

    lock = threading.Lock()
    assert(acquire_lock(lock) is True)  # blocking mode - may wait forever
    assert(acquire_lock(lock, timeout=-1) is False)  # lock is already locked and non-blocking mode is used
    assert(acquire_lock(lock, timeout=1) is False)  # lock is already locked and blocking mode is used
    lock.release()
    assert(acquire_lock(lock, timeout=-1) is True)


@pytest.mark.parametrize(
    "threads_num, increments, lock_decorator", [
        (50, 50, functools.partial(critical_section_dynamic_lock, lock_fn=lambda: threading.Lock()), ),
        (50, 50, functools.partial(critical_section_lock, threading.Lock()), )
    ]
)
def test_critical_section(
    threads_num: int, increments: int, lock_decorator: typing.Callable[..., typing.Callable[..., typing.Any]]
) -> None:

    counter = 0

    def shared_counter_fn() -> None:
        nonlocal counter
        for i in range(increments):
            counter += 1

    thread_safe_fn = lock_decorator()(shared_counter_fn)

    threads = [threading.Thread(target=thread_safe_fn) for _ in range(threads_num)]
    for th in threads:
        th.start()

    for th in threads:
        th.join()

    assert(counter == (threads_num * increments))


def test_critical_section_timeout() -> None:
    lock = threading.Lock()
    lock.acquire()

    def fixed_timeout_fn(
        timeout: typing.Union[int, float, None]
    ) -> typing.Callable[..., typing.Union[int, float, None]]:
        def result(*args: typing.Any, **kwargs: typing.Any) -> typing.Union[int, float, None]:
            return timeout
        return result

    thread_dynamic_safe_fn = critical_section_dynamic_lock(lambda: lock, timeout_fn=fixed_timeout_fn(3))(lambda: None)
    with pytest.raises(CriticalSectionError):
        thread_dynamic_safe_fn()

    thread_safe_fn = critical_section_lock(lock, timeout=3)(lambda: None)
    with pytest.raises(CriticalSectionError):
        thread_safe_fn()

    lock.release()
    thread_dynamic_safe_fn()
    thread_safe_fn()


class SharedResourceSample(CriticalResource):

    def __init__(self) -> None:
        CriticalResource.__init__(self, timeout=3)
        self.counter = 0

    @CriticalResource.critical_section
    def increase(self) -> None:
        self.counter += 1


class TestCriticalResource:

    @pytest.mark.parametrize("threads_num, repeats", [(50, 50)])
    def test(self, threads_num: int, repeats: int) -> None:
        sr = SharedResourceSample()
        assert(sr.counter == 0)

        def thread_fn_increase() -> None:
            for _ in range(repeats):
                sr.increase()

        threads = [threading.Thread(target=thread_fn_increase) for _ in range(threads_num)]
        for th in threads:
            th.start()

        for th in threads:
            th.join()

        assert(sr.counter == (threads_num * repeats))

    def test_exceptions(self) -> None:
        class A:
            @CriticalResource.critical_section
            def foo(self) -> None:
                pass
        with pytest.raises(TypeError):
            A().foo()


class TestLockFreeContext:

    @pytest.mark.parametrize("threads_num, repeats", [(50, 50)])
    def test(self, threads_num: int, repeats: int) -> None:
        sr = SharedResourceSample()
        assert(sr.counter == 0)

        def thread_fn_increase() -> None:
            for i in range(repeats):
                with sr.critical_context(timeout=1) as c1:
                    c1.increase()

        threads = [threading.Thread(target=thread_fn_increase) for x in range(threads_num)]
        for th in threads:
            th.start()

        for th in threads:
            th.join()

        assert(sr.counter == (threads_num * repeats))

        class A(CriticalResource):
            @classmethod
            def foo(cls, i: int) -> int:
                return 2 * i

        a = A()
        with a.critical_context() as c2:
            assert(c2.foo(1) == 2)  # access to original attributes works still

    def test_exceptions(self) -> None:
        sr = SharedResourceSample()
        c1 = sr.critical_context(timeout=1)

        with c1:
            pass  # this is ok
        with pytest.raises(CriticalSectionError):
            with c1:
                with c1:  # there is no way to lock a context twice
                    pass

        c2 = sr.critical_context(timeout=1)
        with c1:
            pass  # this is fine still
        with pytest.raises(CriticalSectionError):
            with c1:
                with c2:  # there is no way to get in a critical section even from different contexts
                    pass

        with c1:
            pass  # this doesn't do anything bad

        with pytest.raises(CriticalSectionError):
            with c1:
                sr.increase()  # this won't work either

        c1.__enter__()  # the object is "locked"
        with pytest.raises(CriticalSectionError):
            sr.increase()  # the object is locked definitely
        del c1
        gc.collect()
        sr.increase()  # the object was unlocked automatically

        with pytest.raises(CriticalSectionError):
            _ = c2.increase  # access to underlined methods are not allowed without a lock



# -*- coding: utf-8 -*-

import itertools

import pytest

from pyknic.lib.registry import NoSuchAPIIdError, DuplicateAPIIdError, APIRegistryProto, APIRegistry, register_api
from pyknic.lib.registry import hash_id_by_tokens


def test_exceptions() -> None:
    assert(issubclass(NoSuchAPIIdError, Exception) is True)
    assert(issubclass(DuplicateAPIIdError, Exception) is True)


def test_abstract() -> None:
    pytest.raises(TypeError, APIRegistryProto)
    pytest.raises(NotImplementedError, APIRegistryProto.register, None, None, None)
    pytest.raises(NotImplementedError, APIRegistryProto.unregister, None, None)
    pytest.raises(NotImplementedError, APIRegistryProto.get, None, None)
    pytest.raises(NotImplementedError, APIRegistryProto.ids, None)
    pytest.raises(NotImplementedError, APIRegistryProto.has, None, None)
    pytest.raises(NotImplementedError, APIRegistryProto.__iter__, None)


class TestAPIRegistry:

    def test(self) -> None:
        registry = APIRegistry()
        pytest.raises(NoSuchAPIIdError, registry.get, 'foo')
        pytest.raises(NoSuchAPIIdError, registry.get, 'bar')

        registry.register('foo', 1)
        assert(registry.get('foo') == 1)
        pytest.raises(NoSuchAPIIdError, registry.get, 'bar')

        assert(registry.has('foo') is True)
        assert('foo' in registry)

        assert(registry.has('bar') is False)
        assert('bar' not in registry)

        pytest.raises(DuplicateAPIIdError, registry.register, 'foo', 1)

        registry.register('bar', 1)
        assert(registry['foo'] == 1)
        assert(registry['bar'] == 1)

        assert(registry.has('foo') is True)
        assert('foo' in registry)

        assert(registry.has('bar') is True)
        assert('bar' in registry)

        secondary_registry = APIRegistry(fallback_registry=registry)
        assert(secondary_registry['foo'] == 1)
        assert(secondary_registry['bar'] == 1)
        pytest.raises(NoSuchAPIIdError, secondary_registry.get, 'zzz')
        pytest.raises(NoSuchAPIIdError, secondary_registry.get, 'xxx')

        registry.register('zzz', 1)
        assert(secondary_registry['foo'] == 1)
        assert(secondary_registry['bar'] == 1)
        assert(secondary_registry['zzz'] == 1)
        pytest.raises(NoSuchAPIIdError, secondary_registry.get, 'xxx')

        secondary_registry.register('xxx', 1)
        assert(secondary_registry['foo'] == 1)
        assert(secondary_registry['bar'] == 1)
        assert(secondary_registry['zzz'] == 1)
        assert(secondary_registry['xxx'] == 1)
        pytest.raises(NoSuchAPIIdError, registry.get, 'xxx')

        secondary_registry.register('zzz', 2)
        assert(secondary_registry['foo'] == 1)
        assert(secondary_registry['bar'] == 1)
        assert(secondary_registry['zzz'] == 2)
        assert(secondary_registry['xxx'] == 1)
        pytest.raises(NoSuchAPIIdError, registry.get, 'xxx')

        registry.unregister('zzz')
        assert(secondary_registry['foo'] == 1)
        assert(secondary_registry['bar'] == 1)
        assert(secondary_registry['xxx'] == 1)
        pytest.raises(NoSuchAPIIdError, registry.get, 'xxx')
        pytest.raises(NoSuchAPIIdError, registry.get, 'zzz')

        pytest.raises(NoSuchAPIIdError, registry.unregister, 'zzz')

        registry_ids_gen = tuple(registry.ids())
        assert(len(registry_ids_gen) == 2)
        assert('foo' in registry_ids_gen)
        assert('bar' in registry_ids_gen)

        secondary_registry_ids_gen = tuple(secondary_registry.ids())
        assert(len(secondary_registry_ids_gen) == 2)
        assert('xxx' in secondary_registry_ids_gen)
        assert('zzz' in secondary_registry_ids_gen)

    def test_iter(self) -> None:
        registry = APIRegistry()

        assert(list(registry) == [])

        registry.register('foo', 1)
        registry.register('bar', 2)
        registry.register('xxx', 3)

        assert(set(registry) == {('foo', 1), ('bar', 2), ('xxx', 3)})


def test_register_api() -> None:

    def foo(a: int, b: int) -> int:
        return a + b

    def bar(a: int, b: int) -> int:
        return a - b

    registry = APIRegistry()

    decorated_foo = register_api(registry)(foo)
    decorated_bar = register_api(registry, api_id='zzz')(bar)

    assert(decorated_foo(3, 4) == 7)
    assert(decorated_bar(3, 4) == -1)

    assert(registry['test_register_api.<locals>.foo'](5, 2) == 7)
    assert(registry['zzz'](5, 2) == 3)

    pytest.raises(NoSuchAPIIdError, registry.get, 'bar')

    class C:
        api_id = 'bar'

    register_api(registry, api_id=lambda x: x.api_id, callable_api_id=True)(C)
    assert(registry['bar'] is C)

    with pytest.raises(ValueError):
        class D:
            pass
        register_api(registry, api_id=1, callable_api_id=True)(D)


def test_hash_id_by_tokens() -> None:
    unsorted_hashes = [
        hash_id_by_tokens('single_token'),
        hash_id_by_tokens('first_token'),
        hash_id_by_tokens('first_token', 'second_token'),
        hash_id_by_tokens('order1', 'order2'),
        hash_id_by_tokens('order2', 'order1'),
        hash_id_by_tokens('order1', 'order2', 'order3'),
    ]

    for hash_a, hash_b in itertools.permutations(unsorted_hashes, 2):
        assert(hash_a != hash_b)

    unsorted_hashes = [
        hash_id_by_tokens('single_token', pre_sort=False),
        hash_id_by_tokens('first_token', pre_sort=False),
        hash_id_by_tokens('first_token', 'second_token', pre_sort=False),
        hash_id_by_tokens('order1', 'order2', pre_sort=False),
        hash_id_by_tokens('order2', 'order1', pre_sort=False),
        hash_id_by_tokens('order1', 'order2', 'order3', pre_sort=False),
    ]

    for hash_a, hash_b in itertools.permutations(unsorted_hashes, 2):
        assert(hash_a != hash_b)

    sorted_hashes = [
        hash_id_by_tokens('single_token', pre_sort=True),
        hash_id_by_tokens('first_token', pre_sort=True),
        hash_id_by_tokens('first_token', 'second_token', pre_sort=True),
        hash_id_by_tokens('order1', 'order2', pre_sort=True),
        hash_id_by_tokens('order1', 'order2', 'order3', pre_sort=True),
    ]

    for hash_a, hash_b in itertools.permutations(sorted_hashes, 2):
        assert(hash_a != hash_b)

    assert(hash_id_by_tokens('order1', 'order2', pre_sort=True) == hash_id_by_tokens('order2', 'order1', pre_sort=True))



# -*- coding: utf-8 -*-

import io

import pytest

from pyknic.lib.config import ConfigStorageProto, Config, ConfigList, ConfigOption

from pyknic.lib.capability import CapabilitiesHolder, iscapable


class TestConfigStorageProto:

    def test(self) -> None:
        storage = ConfigStorageProto()
        assert(isinstance(storage, CapabilitiesHolder))

    def test_exceptions(self) -> None:
        storage = ConfigStorageProto()

        with pytest.raises(NotImplementedError):
            assert(iscapable(storage, ConfigStorageProto.getitem) is False)
            storage[1]

        assert(iscapable(storage, ConfigStorageProto.as_int) is False)
        assert(iscapable(storage, ConfigStorageProto.as_float) is False)
        assert(iscapable(storage, ConfigStorageProto.as_bool) is False)
        assert(iscapable(storage, ConfigStorageProto.as_str) is False)
        for c in (int, float, bool, str):
            with pytest.raises(NotImplementedError):
                c(storage)

        for m in (
                ConfigStorageProto.is_none,
                ConfigStorageProto.properties,
                ConfigStorageProto.reset_properties,
                ConfigStorageProto.iterate_list
        ):
            with pytest.raises(NotImplementedError):
                assert(iscapable(storage, m) is False)
                m(storage)

        for m in (ConfigStorageProto.has_property, ConfigStorageProto.property):  # type: ignore[assignment]
            with pytest.raises(NotImplementedError):
                assert(iscapable(storage, m) is False)
                m(storage, 'foo')   # type: ignore[call-arg]

        with pytest.raises(NotImplementedError):
            assert(iscapable(storage, ConfigStorageProto.merge_file) is False)
            config_file = io.StringIO("")
            storage.merge_file(config_file)

        with pytest.raises(NotImplementedError):
            assert(iscapable(storage, ConfigStorageProto.merge_config) is False)
            storage.merge_config("")  # type: ignore[arg-type]


class TestConfig:

    sample_config = """
section1:
  integer_option: 1

section2:
  float_option: 0.5

section3:
  emoji = 😃

section4:
  - 1
  -
  - foo
  - foo:
    - xxx
    - yyy
    - zzz

big_section:
  general_int: 0
  boolean_value: True
  yes_val: no
  null_val:
  str_val: foo

  prefix:
    int_value: 1
    float_value: 1.2

"""

    extra_config = """
section1:
  integer_option: 2

section5:
  value: foo
"""

    def test_plain(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)

        config = Config()
        assert(config.properties() == set())
        assert(isinstance(config, ConfigStorageProto))

        config = Config(file_obj=config_file)
        assert(config.properties() == {"section1", "section2", "section3", "section4", "big_section"})

        config_file.seek(0)
        config = Config(file_obj=config_file, property_name="section1")
        assert(config.properties() == {"section1", })

    def test_sections(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)
        assert(config.properties() == {"section1", "section2", "section3", "section4", "big_section"})
        assert(config.has_property("section1") is True)
        assert(config.has_property("section5") is False)

    def test_reset(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)
        assert(config.properties() == {"section1", "section2", "section3", "section4", "big_section"})
        config.reset_properties()
        assert(config.properties() == set())

    def test_section(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)

        section = config['big_section']
        assert(isinstance(section, Config))
        assert(section.has_property('general_int') is True)
        assert(section.has_property('prefix') is True)
        assert(section.has_property('int_value') is False)
        assert(section.has_property('float_value') is False)

        section = config.property('big_section')
        assert(isinstance(section, Config))
        assert(section.has_property('general_int') is True)
        assert(section.has_property('prefix') is True)
        assert(section.has_property('int_value') is False)
        assert(section.has_property('float_value') is False)

        sub_section = section['prefix']
        assert(isinstance(sub_section, Config))
        assert(sub_section.has_property('general_int') is False)
        assert(sub_section.has_property('prefix') is False)
        assert(sub_section.has_property('int_value') is True)
        assert(sub_section.has_property('float_value') is True)

    def test_merge_file(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)
        assert(config.has_property('section5') is False)
        assert(int(config['section1']['integer_option']) == 1)

        extra_file = io.StringIO(TestConfig.extra_config)
        config.merge_file(extra_file)
        assert(config.has_property('section5') is True)
        assert(int(config['section1']['integer_option']) == 2)

    def test_config_merge(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)
        assert(config.has_property('section5') is False)
        assert(int(config['section1']['integer_option']) == 1)

        extra_file = io.StringIO(TestConfig.extra_config)
        extra_config = Config(file_obj=extra_file)
        config.merge_config(extra_config)
        assert(config.has_property('section5') is True)
        assert(int(config['section1']['integer_option']) == 2)

    def test_partial_config_merge(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)

        extra_file = io.StringIO(TestConfig.extra_config)
        extra_config = Config(file_obj=extra_file)
        config.merge_config(extra_config, property_name='section1')

        assert(config.has_property('section5') is False)
        assert(int(config['section1']['integer_option']) == 2)

    def test_exceptions(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)

        with pytest.raises(TypeError):
            config[0]

        with pytest.raises(ValueError):
            config.property('unknown_section')

        with pytest.raises(ValueError):
            config['unknown_section']

        extra_file = io.StringIO(TestConfig.extra_config)
        extra_config = Config(file_obj=extra_file)

        with pytest.raises(ValueError):
            config.merge_config(extra_config, 'unknown_section')

        with pytest.raises(TypeError):
            config.merge_config(ConfigList())


class TestConfigList:

    def test(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)

        config_list = config['section4']
        assert(isinstance(config_list, ConfigList))
        assert(isinstance(config_list, ConfigStorageProto))

        config_list_items = list(config_list.iterate_list())
        assert(len(config_list_items) == 4)

        assert(isinstance(config_list[0], ConfigOption))
        assert(config_list[0].is_none() is False)
        assert(int(config_list[0]) == 1)

        assert(isinstance(config_list[1], ConfigOption))
        assert(config_list[1].is_none() is True)

        assert(isinstance(config_list[2], ConfigOption))
        assert(config_list[2].is_none() is False)
        assert(str(config_list[2]) == "foo")

        assert(isinstance(config_list[3], ConfigOption) is False)
        assert(isinstance(config_list[3], Config))
        assert(config_list[3].properties() == {"foo", })
        assert([str(x) for x in config_list[3]["foo"].iterate_list()] == ["xxx", "yyy", "zzz"])


class TestConfigOption:

    def test(self) -> None:
        config_file = io.StringIO(TestConfig.sample_config)
        config = Config(file_obj=config_file)

        option = config['section1']['integer_option']
        assert(isinstance(option, ConfigOption))
        assert(isinstance(option, ConfigStorageProto))
        pytest.raises(TypeError, str, option)
        assert(int(option) == 1)

        assert(float(config['section2']['float_option']) == 0.5)
        assert(bool(config['big_section']['boolean_value']) is True)
        assert(bool(config['big_section']['yes_val']) is False)
        assert(config['big_section']['null_val'].is_none())
        assert(str(config['big_section']['str_val']) == 'foo')




import asyncio
import io
import typing
import warnings

import decorator
import fastapi
import uvicorn
import pytest

from pyknic.lib.fastapi.base import BaseFastAPIApp
from pyknic.path import root_path
from pyknic.lib.config import Config
from pyknic.lib.gettext import GetTextWrapper

from fixtures.asyncio import BaseAsyncFixture
from fixtures.event_loop import EventLoopDescriptor
from fixture_helpers import pyknic_fixture


class AsyncFastAPIFixture(BaseAsyncFixture):

    __startup_retries__ = 10
    __startup_pause__ = 0.5

    def __init__(self) -> None:
        BaseAsyncFixture.__init__(self)
        self.fastapi = fastapi.FastAPI()

        self.config = uvicorn.Config(self.fastapi)
        self.server = uvicorn.Server(self.config)

        with open(root_path / 'tasks/fastapi/config.yaml') as f:
            self.default_config = Config(file_obj=f)
        self.gettext = GetTextWrapper(root_path / 'locales')
        self.configured_with: typing.Optional[typing.Type[BaseFastAPIApp]] = None
        self.app_config = Config()

    async def start_async_service(self, loop_descriptor: EventLoopDescriptor) -> None:
        assert(not self.server.started)
        if not self.server.started:

            with warnings.catch_warnings():
                # TODO: check deprecation that uvicorn produce
                warnings.filterwarnings("ignore", category=DeprecationWarning, message="websockets.legacy")
                warnings.filterwarnings(
                    "ignore", category=DeprecationWarning, message="websockets.server.WebSocketServerProtocol"
                )
                await self.server.serve()

    async def wait_startup(self, loop_descriptor: EventLoopDescriptor) -> None:
        for _ in range(self.__startup_retries__):
            if self.server.started:
                break
            await asyncio.sleep(self.__startup_pause__)
        assert(self.server.started)

    async def flush_async(self, loop_descriptor: EventLoopDescriptor) -> None:
        self.fastapi.routes.clear()
        self.configured_with = None

    async def stop_async(self) -> None:
        self.server.should_exit = True
        await self.server.shutdown()

    def setup_fastapi(
        self, fast_api_cls: typing.Type[BaseFastAPIApp], extra_config: str | None = None
    ) -> None:
        if self.configured_with is None:
            self.configured_with = fast_api_cls

            self.app_config = Config()
            self.app_config.merge_config(self.default_config)
            if extra_config:
                self.app_config.merge_config(Config(file_obj=io.StringIO(extra_config)))

            fast_api_cls.create_app(self.fastapi, self.app_config, self.gettext)
        elif self.configured_with is not fast_api_cls:
            raise ValueError('Already configured fixture!')

    @staticmethod
    def base_config(
        fast_api_cls: typing.Type[BaseFastAPIApp],
        extra_config: str | None = None,
    ) -> typing.Callable[..., typing.Any]:

        def first_level_decorator(
            decorated_function: typing.Callable[..., typing.Any]
        ) -> typing.Callable[..., typing.Any]:
            def second_level_decorator(
                original_function: typing.Callable[..., typing.Any], *args: typing.Any, **kwargs: typing.Any
            ) -> typing.Any:

                fixture_found = False
                for i in args:
                    if isinstance(i, AsyncFastAPIFixture):
                        if fixture_found:
                            raise RuntimeError('Multiple AsyncFastAPIFixture instances found!')

                        fixture_found = True
                        i.setup_fastapi(fast_api_cls, extra_config=extra_config)

                if not fixture_found:
                    raise RuntimeError('No suitable fixture found')

                return original_function(*args, **kwargs)

            return decorator.decorator(  # type: ignore[no-any-return, no-untyped-call]
                second_level_decorator
            )(decorated_function)
        return first_level_decorator


def _fastapi_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from pyknic_fixture(AsyncFastAPIFixture)


@pytest.fixture
def fastapi_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from _fastapi_fixture()


@pytest.fixture(scope='class')
def fastapi_class_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from _fastapi_fixture()


@pytest.fixture(scope='module')
def fastapi_module_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from _fastapi_fixture()


@pytest.fixture(scope='package')
def fastapi_package_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from _fastapi_fixture()


@pytest.fixture(scope='session')
def fastapi_session_fixture() -> typing.Generator[AsyncFastAPIFixture, None, None]:
    yield from _fastapi_fixture()



# -*- coding: utf-8 -*-
# pyknic/lib/registry.py
#
# Copyright (C) 2019-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import typing

from abc import ABCMeta, abstractmethod

from pyknic.lib.verify import verify_value


class NoSuchAPIIdError(Exception):
    """ This exception is raised when a looked up API id is not found
    """
    pass


class DuplicateAPIIdError(Exception):
    """ This exception is raised when an attempt to register a descriptor with an id, that has been already
    registered, is made
    """
    pass


class APIRegistryProto(metaclass=ABCMeta):
    """ This is prototype for a general registry, object that stores anything by id. It may look like a dict object,
    but this class should be used in order to distinguish registry operation and commonly used dict
   """

    @abstractmethod
    def register(self, api_id: typing.Hashable, api_descriptor: typing.Any) -> None:
        """ Save the specified descriptor by the specified id

        :param api_id: unique id by which a descriptor may be found
        :param api_descriptor: descriptor that should be stored in this registry

        :raise DuplicateAPIIdError: when the specified API id has been registered already
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def unregister(self, api_id: typing.Hashable) -> None:
        """ Remove the specified descriptor from registry

        :param api_id: id that will be removed from this registry

        :raise NoSuchAPIIdError: when the specified API id has not been registered
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def get(self, api_id: typing.Hashable) -> typing.Any:
        """ Retrieve previously saved descriptor by an id

        :param api_id: id of a target descriptor

        :raise NoSuchAPIIdError: when the specified API id has not been registered
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def has(self, api_id: typing.Hashable) -> bool:
        """ Check if this registry has the specified id

        :param api_id: id to check
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def ids(self) -> typing.Generator[typing.Hashable, None, None]:
        """ Return generator that will return all ids that this registry have
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def __iter__(self) -> typing.Generator[typing.Tuple[typing.Hashable, typing.Any], None, None]:
        """ Iterate over all the registered descriptors
        """
        raise NotImplementedError('This method is abstract')


class APIRegistry(APIRegistryProto):
    """ This is a basic registry implementation. It behaves like a dict mostly
    """

    def __init__(self, fallback_registry: typing.Optional[APIRegistryProto] = None):
        """ Create new registry

        :param fallback_registry: a registry where entries will be looked up if they are not found in
        this registry. This parameter helps to use all the items that are registered in other registry
        without registrations repeat
        """
        APIRegistryProto.__init__(self)
        self.__descriptors: typing.Dict[typing.Hashable, typing.Any] = dict()
        self.__fallback_registry = fallback_registry

    def register(self, api_id: typing.Hashable, api_descriptor: typing.Any) -> None:
        """ :meth:`.APIRegistryProto.register` method implementation
        """
        if api_id in self.__descriptors:
            raise DuplicateAPIIdError('The specified id "%s" has been used already' % str(api_id))
        self.__descriptors[api_id] = api_descriptor

    def unregister(self, api_id: typing.Hashable) -> None:
        """ :meth:`.APIRegistryProto.unregister` method implementation
        """
        if api_id not in self.__descriptors:
            raise NoSuchAPIIdError('No such entry: %s' % api_id)
        del self.__descriptors[api_id]

    def get(self, api_id: typing.Hashable) -> typing.Any:
        """ :meth:`.APIRegistryProto.register` method implementation
        """
        try:
            return self.__descriptors[api_id]
        except KeyError:
            pass

        if self.__fallback_registry is not None:
            return self.__fallback_registry.get(api_id)

        raise NoSuchAPIIdError('No such entry: %s' % api_id)

    def __getitem__(self, item: typing.Hashable) -> typing.Any:
        """ Shortcut to :meth:`.APIRegistryProto.get`
        """
        return self.get(item)

    def ids(self) -> typing.Generator[typing.Hashable, None, None]:
        """ :meth:`.APIRegistryProto.ids` method implementation
        """
        return (x for x in self.__descriptors.keys())

    def has(self, api_id: typing.Hashable) -> bool:
        """ :meth:`.APIRegistryProto.has` method implementation
        """
        return api_id in self.__descriptors

    def __contains__(self, item: typing.Hashable) -> bool:
        """ Shortcut to :meth:`.APIRegistryProto.has`
        """
        return self.has(item)

    def __iter__(self) -> typing.Generator[typing.Tuple[typing.Hashable, typing.Any], None, None]:
        """ :meth:`.APIRegistryProto.__iter__` method implementation
        """
        yield from self.__descriptors.items()


def register_api(
    registry: APIRegistryProto,
    api_id: typing.Optional[
        typing.Union[typing.Hashable, typing.Callable[[typing.Any], typing.Hashable]]
    ] = None,
    callable_api_id: bool = False
) -> typing.Callable[..., typing.Callable[..., typing.Any]]:
    """ This decorator helps to register function, static method or class in the specified registry

    :param registry: registry to which a function should be registered
    :param api_id: id with which function will be registered. If it is not specified then function qualification \
    name will be used
    :param callable_api_id: whether 'api_id' is not an entry identifier but a callable (function) that accepts
    decorated object in order to retrieve a real id

    :raise ValueError: if the 'callable_api_id' variable is True but the 'api_id' is not callable object
    """
    def decorator_fn(decorated_obj: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
        nonlocal api_id, callable_api_id  # noqa: F824
        if callable_api_id is True:
            if not callable(api_id):
                raise ValueError('Unable to retrieve an id - "api_id" is non-callable')
            api_id = api_id(decorated_obj)
        elif api_id is None:
            api_id = decorated_obj.__qualname__

        reg_id = api_id
        registry.register(reg_id, decorated_obj)
        return decorated_obj

    return decorator_fn


@verify_value(first_token=lambda x: x.isascii(), other_tokens=lambda x: x.isascii())
def hash_id_by_tokens(first_token: str, *other_tokens: str, pre_sort: bool = False) -> typing.Hashable:
    """Generate a determinate value as api_id by input tokens

    :param first_token: minimal required token to generate a hash
    :param other_tokens: other tokens for hash generation
    :param pre_sort: whether to sort tokens before hashing or not. Different parameters order will lead to different
    results when this value is False
    """

    tokens = [first_token, *other_tokens]

    if pre_sort:
        tokens.sort()

    result = hashlib.sha256()
    for i in tokens:
        result.update(i.encode('ascii'))
    return result.digest()



# -*- coding: utf-8 -*-

import functools
import threading
import typing
import pytest

from dataclasses import dataclass

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SignalsRegistry

from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.datalog.datalog_py import DatalogPy
from pyknic.lib.datalog.datalog import Datalog


class TestDatalog:

    @pytest.mark.parametrize(
        "test_cls", [
            DatalogPy,
            Datalog
        ]
    )
    def test_plain(self, test_cls: typing.Type[DatalogProto]) -> None:
        test_obj = test_cls()
        assert(isinstance(test_obj, DatalogProto) is True)

        seq = [1, True, "foo", ("bar", 0)]
        reversed_seq = seq.copy()
        reversed_seq.reverse()

        for i in seq:
            test_obj.append(i)

        assert(list(test_obj.iterate()) == seq)
        assert(list(test_obj.iterate(reverse=True)) == reversed_seq)

        test_obj.truncate(3)
        result = seq[-3:]
        reversed_result = result.copy()
        reversed_result.reverse()
        assert(list(test_obj.iterate()) == result)
        assert(list(test_obj.iterate(reverse=True)) == reversed_result)

        test_obj.truncate(0)
        assert(list(test_obj.iterate()) == list())
        assert(list(test_obj.iterate(reverse=True)) == list())

    @pytest.mark.parametrize(
        "test_cls", [
            DatalogPy,
            Datalog
        ]
    )
    def test_concurrency(self, test_cls: typing.Type[DatalogProto]) -> None:
        test_obj = test_cls()
        start_event = threading.Event()

        def concurrency_push(number_of_object: int) -> None:
            nonlocal test_obj, start_event  # noqa: F824
            start_event.wait()
            for i in range(number_of_object):
                test_obj.append(i)

        threads_number = 20
        number_of_objects = 1000
        threads = [
            threading.Thread(target=functools.partial(concurrency_push, number_of_objects))
            for _ in range(threads_number)
        ]

        for th in threads:
            th.start()

        start_event.set()
        for th in threads:
            th.join()

        result = list(test_obj.iterate())
        result.sort()

        expected_result = ([x for x in range(number_of_objects)] * threads_number)
        expected_result.sort()

        assert(result == expected_result)

    @pytest.mark.parametrize(
        "test_cls", [
            DatalogPy,
            Datalog
        ]
    )
    def test_truncate(self, test_cls: typing.Type[DatalogProto]) -> None:
        test_obj = test_cls()

        for i in range(10):
            test_obj.append(i)

        test_obj.truncate(7)
        assert(list(test_obj.iterate()) == [3, 4, 5, 6, 7, 8, 9])

    @pytest.mark.parametrize(
        "test_cls", [
            DatalogPy,
            Datalog
        ]
    )
    def test_signal(self, test_cls: typing.Type[DatalogProto], signals_registry: 'SignalsRegistry') -> None:
        log = test_cls()

        log.callback(DatalogProto.new_entry, signals_registry)
        assert(signals_registry.dump(True) == [])

        entry_object = (1, "foo", None)
        log.append(entry_object)
        assert(signals_registry.dump(True) == [(log, DatalogProto.new_entry, entry_object)])

    @pytest.mark.parametrize(
        "test_cls", [
            DatalogPy,
            Datalog
        ]
    )
    def test_find(self, test_cls: typing.Type[DatalogProto]) -> None:

        @dataclass
        class LogEntry:
            str_field: str
            int_field: int
            optional_float_field: typing.Optional[float]

        test_obj = test_cls()
        first_record = LogEntry('foo', 10, None)
        second_record = LogEntry('bar', 5, None)
        third_record = LogEntry('other string', 5, 0.1)
        fourth_record = LogEntry('one more string', 12, None)

        test_obj.append(first_record)
        test_obj.append(second_record)
        test_obj.append(third_record)
        test_obj.append(fourth_record)

        assert(test_obj.find(lambda x: x.optional_float_field is None) is first_record)
        assert(test_obj.find(lambda x: x.optional_float_field is None, reverse=True) is fourth_record)

        assert(test_obj.find(lambda x: x.optional_float_field is not None) is third_record)
        assert(test_obj.find(lambda x: x.optional_float_field is not None, reverse=True) is third_record)

        assert(test_obj.find(lambda x: x.int_field == 5) is second_record)
        assert(test_obj.find(lambda x: x.int_field == 5, reverse=True) is third_record)

        assert(test_obj.find(lambda x: x.int_field == 13) is None)
        assert(test_obj.find(lambda x: x.int_field == 13, reverse=True) is None)



# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/login.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import pydantic
import rich.prompt

from pyknic.lib.bellboy.app import register_bellboy_command, BellBoyCommandHandler, LobbyClient
from pyknic.lib.bellboy.models import GeneralBellBoyCommandModel
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyCommandRequest, LobbyStrFeedbackResult
from pyknic.lib.integrated_commands.ping import LobbyPingCommand


@register_bellboy_command()
class LoginCommand(BellBoyCommandHandler):
    """Login to a custom lobby server
    """

    @classmethod
    def command_name(cls) -> str:
        """The :meth:`.BellBoyCommandHandler.command_name` method implementation"""
        return "login"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """The :meth:`.BellBoyCommandHandler.command_model` method implementation"""
        return GeneralBellBoyCommandModel

    async def exec(self) -> LobbyCommandResult:
        """The :meth:`.BellBoyInternalCommand.exec_from_cli` method implementation
        """
        assert(isinstance(self._args, GeneralBellBoyCommandModel))

        secret_token = rich.prompt.Prompt.ask('Secret token', password=True)  # TODO: this looks pretty ugly =(
        # there should be direct interaction with CLI

        server_fingerprint = await LobbyClient.fingerprint(self._args.lobby_url)

        # just to check that auth is ok
        await LobbyClient(self._args.lobby_url, server_fingerprint, secret_token).command_request(
            LobbyCommandRequest(
                name=LobbyPingCommand.command_name(),
                args=LobbyPingCommand.command_model()().model_dump()
            )
        )

        secret_backend = self.secret_backend(self._args.secret_backend)
        secret_backend.set_secret(self._args.lobby_url, server_fingerprint, secret_token)
        return LobbyStrFeedbackResult(str_result=f'Successful login to the {self._args.lobby_url}')



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/queue.py
#
# Copyright (C) 2017-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from datetime import datetime, timezone

from pyknic.lib.signals.proto import Signal
from pyknic.lib.signals.source import SignalSource

from pyknic.lib.tasks.proto import ScheduleRecordProto, ScheduledTaskPostponePolicy


class SchedulerQueue(SignalSource):
    """ This class is a part of :class:`.SchedulerProto` implementation. This class allows to postpone records.

    :note: Methods in this class are not thread safe!
    """

    task_dropped = Signal(ScheduleRecordProto)    # a scheduled task dropped and would not start
    task_postponed = Signal(ScheduleRecordProto)  # a scheduled task postponed and will start later
    task_expired = Signal(ScheduleRecordProto)    # a scheduled task dropped because of expired ttl

    def __init__(self) -> None:
        """ Create a new queue with postpone records
        """
        SignalSource.__init__(self)
        self.__postponed_records: typing.List[ScheduleRecordProto] = []

    def __len__(self) -> int:
        """ Return number of postponed tasks
        """
        return len(self.__postponed_records)

    def __drop_all(self, group_id: str) -> None:
        """ Drop from a queue all record with the specified group_id

        :param group_id: id of a group which records should be dropped
        """
        dropped_records = 0
        for i in range(len(self.__postponed_records)):
            check_record = self.__postponed_records[i - dropped_records]
            if check_record.group_id() == group_id:
                self.__postponed_records.pop(i - dropped_records)
                dropped_records += 1
                self.emit(self.task_dropped, check_record)

    def __keep_first(self, group_id: str) -> bool:
        """ Try to keep the earliest record of the same group. Other records of the same group will be dropped

        :param group_id: id of a group which records should be dropped (all but the earliest)

        :return: return True if the earliest record was found and return False otherwise
        """
        first_found = False
        dropped_records = 0
        for i in range(len(self.__postponed_records)):
            check_record = self.__postponed_records[i - dropped_records]
            if check_record.group_id() == group_id:
                if first_found:
                    self.__postponed_records.pop(i - dropped_records)
                    self.emit(self.task_dropped, check_record)
                    dropped_records += 1
                else:
                    first_found = True
        return first_found

    def postpone(self, record: ScheduleRecordProto) -> None:
        """ Postpone a record (or drop it because of policy and/or ttl)

        :param record: record to postpone
        """
        ttl = record.ttl()
        group_id = record.group_id()
        postpone_policy = record.postpone_policy()

        if postpone_policy == ScheduledTaskPostponePolicy.drop:
            self.emit(self.task_dropped, record)
            return

        if ttl is not None and ttl < datetime.now(timezone.utc).timestamp():
            self.emit(self.task_expired, record)
            return

        if postpone_policy == ScheduledTaskPostponePolicy.wait or group_id is None:
            self.__postponed_records.append(record)
            self.emit(self.task_postponed, record)
            return

        if postpone_policy == ScheduledTaskPostponePolicy.keep_last:
            self.__drop_all(group_id)
            self.__postponed_records.append(record)
            self.emit(self.task_postponed, record)
            return

        if postpone_policy == ScheduledTaskPostponePolicy.keep_first:
            if not self.__keep_first(group_id):
                self.__postponed_records.append(record)
                self.emit(self.task_postponed, record)
            else:
                self.emit(self.task_dropped, record)

    def next_record(
        self,
        filter_fn: typing.Optional[typing.Callable[[ScheduleRecordProto], bool]] = None
    ) -> typing.Optional[ScheduleRecordProto]:
        """ Get record from a queue to be executed next. The earliest records win, but not always

        :param filter_fn: check a record with this function. If record is suitable then this function should return
        True or return False otherwise

        :return: return a record that should be executed next or return None if no suitable record is found
        """

        utc_now = datetime.now(timezone.utc).timestamp()
        dropped_records = 0

        for i in range(len(self.__postponed_records)):
            record = self.__postponed_records[i - dropped_records]

            ttl = record.ttl()
            if ttl is not None and ttl < utc_now:
                self.__postponed_records.pop(i - dropped_records)
                self.emit(self.task_expired, record)
                dropped_records += 1
                continue

            if filter_fn and not filter_fn(record):
                continue

            self.__postponed_records.pop(i - dropped_records)
            return record
        return None



# -*- coding: utf-8 -*-
# pyknic/lib/singleton.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import ABCMeta


class SingletonMeta(ABCMeta):
    """ This metaclass helps to implement a transparent globally accepted instance (singleton)
    """

    __instance__ = None  # singleton instance

    def __init__(cls, name: str, bases: typing.Tuple[type], namespace: typing.Dict[str, typing.Any]):
        """ Create a new instance
        """
        ABCMeta.__init__(cls, name, bases, namespace)

    def __getattribute__(cls, item: str) -> typing.Any:
        """ Try to get attribute from an instance
        """
        if item == "__instance__":
            return ABCMeta.__getattribute__(cls, item)

        if cls.__instance__ is not None and item in dir(cls.__instance__):
            return cls.__instance__.__getattribute__(item)

        return ABCMeta.__getattribute__(cls, item)

    def __setattr__(cls, key: str, value: typing.Any) -> None:
        """ Try to set attribute for an instance
        """
        if cls.__instance__ is not None and key in dir(cls.__instance__):
            cls.__instance__.__setattr__(key, value)
        else:
            ABCMeta.__setattr__(cls, key, value)

    def setup_singleton(cls, value: typing.Any) -> None:
        """ Setup an instance as a singleton. Each instance may be set up only once
        """
        if cls.__instance__ is not None:
            raise ValueError('Singleton has been initialized already')
        cls.__instance__ = value

    def singleton(cls) -> typing.Any:
        """ Return an instance or None if an instance has not been set up
        """
        return cls.__instance__


def create_singleton(value: typing.Any) -> type:
    """ Return initialized singleton class

    :param value: value to set up a singleton
    """
    class Singleton(metaclass=SingletonMeta):
        __instance__ = value
    return Singleton



# -*- coding: utf-8 -*-

import functools
import pytest
import typing

from decorator import decorator

from pyknic.lib.capability import CapabilityDescriptor, capability, CapabilitiesHolderMeta, iscapable
from pyknic.lib.capability import CapabilitiesHolder

from pyknic.lib.x_mansion import CapabilitiesAndSignals

from pyknic.lib.typing import GenericFunc, P, R


class TestCapabilityDescriptor:

    def test(self) -> None:
        class A:
            pass

        d = CapabilityDescriptor(A, 'foo')
        assert(d.cls() is A)
        assert(d.name() == 'foo')


class TestCapabilitiesHolderMeta:

    @staticmethod
    def dumb_decorator(f: GenericFunc[P, R]) -> GenericFunc[P, R]:
        def decorator_fn(f: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            return f(*args, **kwargs)

        return decorator(decorator_fn)(f)  # type: ignore[no-any-return, no-untyped-call]

    def test(self) -> None:

        class A(metaclass=CapabilitiesHolderMeta):

            @capability
            def foo(self) -> int:
                return 1

            def bar(self) -> int:
                return 2

        with pytest.raises(TypeError):
            class B(metaclass=CapabilitiesHolderMeta):
                @classmethod
                @capability
                def foo(cls) -> int:
                    return 1

    def test_iscapable(self) -> None:
        class A(metaclass=CapabilitiesHolderMeta):

            @TestCapabilitiesHolderMeta.dumb_decorator
            @capability
            def foo(self) -> int:
                return 1

            @capability
            @TestCapabilitiesHolderMeta.dumb_decorator
            def bar(self) -> int:
                return 2

            def zzz(self) -> int:
                return 3

        class B(A):

            def foo(self) -> int:
                return 4

        class C(A):

            def bar(self) -> int:
                return 5

        class D(A):

            def foo(self) -> int:
                return 6

            def bar(self) -> int:
                return 7

        a = A()
        b = B()
        c = C()
        d = D()

        assert(a.foo() == 1)
        assert(a.bar() == 2)
        assert(a.zzz() == 3)

        assert(b.foo() == 4)
        assert(b.bar() == 2)
        assert(b.zzz() == 3)

        assert(c.foo() == 1)
        assert(c.bar() == 5)
        assert(c.zzz() == 3)

        assert(d.foo() == 6)
        assert(d.bar() == 7)
        assert(d.zzz() == 3)

        assert(iscapable(A, A.foo) is False)
        assert(iscapable(A, A.bar) is False)
        assert(iscapable(A, B.bar) is False)
        assert(iscapable(A, C.foo) is False)

        assert(iscapable(a, A.foo) is False)
        assert(iscapable(a, A.bar) is False)
        assert(iscapable(a, B.bar) is False)
        assert(iscapable(a, C.foo) is False)

        assert(iscapable(B, A.foo) is True)
        assert(iscapable(B, A.bar) is False)
        assert(iscapable(B, B.bar) is False)
        assert(iscapable(B, C.foo) is True)

        assert(iscapable(b, A.foo) is True)
        assert(iscapable(b, A.bar) is False)
        assert(iscapable(b, B.bar) is False)
        assert(iscapable(b, C.foo) is True)

        assert(iscapable(C, A.foo) is False)
        assert(iscapable(C, A.bar) is True)
        assert(iscapable(C, B.bar) is True)
        assert(iscapable(C, C.foo) is False)

        assert(iscapable(c, A.foo) is False)
        assert(iscapable(c, A.bar) is True)
        assert(iscapable(c, B.bar) is True)
        assert(iscapable(c, C.foo) is False)

        assert(iscapable(D, A.foo) is True)
        assert(iscapable(D, A.bar) is True)
        assert(iscapable(D, B.bar) is True)
        assert(iscapable(D, C.foo) is True)

        assert(iscapable(d, A.foo) is True)
        assert(iscapable(d, A.bar) is True)
        assert(iscapable(d, B.bar) is True)
        assert(iscapable(d, C.foo) is True)


class TestCapabilitiesHolder:

    @pytest.mark.parametrize(
        "test_cls", [
            CapabilitiesHolder,
            CapabilitiesAndSignals,
        ]
    )
    def test(self, test_cls: typing.Type[CapabilitiesHolder]) -> None:
        assert(isinstance(test_cls, CapabilitiesHolderMeta) is True)

        class A(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            @capability
            def foo(self) -> None:
                pass

            @capability
            def bar(self) -> None:
                pass

        class B(A):
            pass

        class C(A):

            def foo(self) -> None:
                pass

        a = A()
        b = B()
        c = C()

        assert(iscapable(a, A.foo) is False)
        assert(iscapable(b, A.foo) is False)
        assert(iscapable(c, A.foo) is True)

        assert(iscapable(A, A.foo) is False)
        assert(iscapable(B, A.foo) is False)
        assert(iscapable(C, A.foo) is True)

        assert(iscapable(a, A.bar) is False)
        assert(iscapable(b, A.bar) is False)
        assert(iscapable(c, A.bar) is False)

        assert(iscapable(A, A.bar) is False)
        assert(iscapable(B, A.bar) is False)
        assert(iscapable(C, A.bar) is False)

        assert(A.foo not in a)
        assert(A.foo not in b)
        assert(A.foo in c)

        assert(A.bar not in a)
        assert(A.bar not in a)
        assert(A.bar not in a)

        class D(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            @capability
            def foo(self) -> None:
                pass

            @capability
            def bar(self) -> None:
                pass

        assert(iscapable(c, D.foo) is False)
        assert(D.foo not in c)

    @pytest.mark.parametrize(
        "test_cls", [
            CapabilitiesHolder,
            CapabilitiesAndSignals,
        ]
    )
    def test_dynamic_capability(self, test_cls: typing.Type[CapabilitiesHolder]) -> None:

        class A(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            @capability
            def foo(self, check_obj: typing.Any) -> None:
                pass

        assert(iscapable(A(), A.foo) is False)

        class B(A):
            pass

        b = B()
        assert(iscapable(b, A.foo) is False)

        def bar(self: B, check_obj: typing.Any) -> None:
            assert(self is check_obj)  # check that method is bound

        b.append_capability(A.foo, functools.partial(bar, b))
        assert(iscapable(b, A.foo) is True)

        b.foo(b)

    @pytest.mark.parametrize(
        "test_cls", [
            CapabilitiesHolder,
            CapabilitiesAndSignals,
        ]
    )
    def test_dynamic_capability_exceptions(self, test_cls: typing.Type[CapabilitiesHolder]) -> None:

        class A(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            @capability
            def foo(self) -> None:
                pass

        class B(A):
            pass

        class C(A):
            def foo(self) -> None:
                pass

        b = B()
        c = C()

        assert(iscapable(b, A.foo) is False)
        assert(iscapable(c, A.foo) is True)

        def bar() -> None:
            pass

        with pytest.raises(ValueError):
            c.append_capability(A.foo, bar)

        b.append_capability(A.foo, bar)
        assert(iscapable(b, A.foo) is True)

        with pytest.raises(ValueError):
            b.append_capability(A.foo, bar)



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/chain_source.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: more features:
#  - a chained task should register it's result in a datalog
#  - there should be a cooldown procedure for tasks that failed to run from the first time

import enum
import functools
import typing
import uuid

from contextlib import suppress
from datetime import datetime, timezone

from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.datalog.datalog import Datalog
from pyknic.lib.registry import APIRegistryProto, APIRegistry
from pyknic.lib.signals.proto import SignalSourceProto, Signal
from pyknic.lib.signals.proxy import QueueProxy, QueueCallbackException
from pyknic.lib.signals.extra import SignalWaiter, BoundedCallback, ReceivedSignal
from pyknic.lib.tasks.proto import ScheduleSourceProto, TaskProto, TaskResult, ScheduledTaskPostponePolicy
from pyknic.lib.tasks.proto import ScheduleRecordProto
from pyknic.lib.tasks.proto import SchedulerFeedback, SchedulerProto
from pyknic.lib.tasks.scheduler.record import ScheduleRecord


__default_chained_tasks_registry__ = APIRegistry()


@enum.unique
class ChainedTaskState(enum.Enum):
    """ This represents a task state that will be kept in a log
    """
    started = enum.auto()    # a task started
    completed = enum.auto()  # a task finished with result (a task should report this)
    finalized = enum.auto()  # a task has been cleaned up


class ChainedTaskLogEntry:
    """ This is a record in a log, that describes an event happened to a task

    :note: this class was not made with dataclasses because it should be as much "immutable" as possible
    """

    def __init__(
        self,
        api_id: str,
        uid: uuid.UUID,
        state: ChainedTaskState,
        result: typing.Optional[TaskResult] = None
    ):
        """ Create an entry
        """
        self.__api_id = api_id
        self.__uid = uid
        self.__event_datetime = datetime.now(timezone.utc)
        self.__state = state
        self.__result = result

    @property
    def api_id(self) -> str:
        return self.__api_id

    @property
    def uid(self) -> uuid.UUID:
        return self.__uid

    @property
    def event_datetime(self) -> datetime:
        return self.__event_datetime

    @property
    def state(self) -> ChainedTaskState:
        return self.__state

    @property
    def result(self) -> typing.Optional[TaskResult]:
        return self.__result


# noinspection PyAbstractClass
class ChainedTask(TaskProto):
    """ This class may be started by a :class:`.ChainedTasksSource` that respects dependencies
    """

    def __init__(self, datalog: DatalogProto, api_id: str, uid: uuid.UUID):
        TaskProto.__init__(self)
        self.__datalog = datalog
        self.__api_id = api_id
        self.__uid = uid

    def datalog(self) -> DatalogProto:
        return self.__datalog

    def api_id(self) -> str:
        return self.__api_id

    def uid(self) -> uuid.UUID:
        return self.__uid

    @classmethod
    def create(cls, datalog: DatalogProto, api_id: str, uid: uuid.UUID) -> 'ChainedTask':
        """ Create a new task with:

        :param datalog: a log where result should be saved
        :param api_id: an id with which this task will be created
        :param uid: identifier of a task instance
        """
        return cls(datalog, api_id, uid)

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ Return task dependencies
        """
        return None

    def wait_for(self, api_id: str) -> typing.Optional[TaskResult]:
        return ChainedTasksSource.wait_for(self.datalog(), api_id)

    def save_result(self, result: typing.Any) -> None:
        self.datalog().append(
            ChainedTaskLogEntry(self.api_id(), self.uid(), ChainedTaskState.completed, TaskResult(result=result))
        )


class ChainedTasksSource(ScheduleSourceProto, TaskProto):
    """ This is a source for a scheduler that may start tasks and theirs dependencies
    """

    def __init__(
        self,
        datalog: typing.Optional[DatalogProto] = None,
        registry: typing.Optional[APIRegistryProto] = None
    ):
        """ Create a new source

        :param datalog: a log where tasks states are stored
        :param registry: a registry that holds classes (the "__default_chained_tasks_registry__" is used by default)
        """
        ScheduleSourceProto.__init__(self)
        TaskProto.__init__(self)

        self.__queue_proxy = QueueProxy()
        self.__source_uid = str(uuid.uuid4())

        self.__registry = registry if registry else __default_chained_tasks_registry__
        self.__datalog = datalog if datalog else Datalog()

        self.__scheduler: typing.Optional[SchedulerProto] = None
        self.__record_completed_clbk = BoundedCallback(self.__record_completed)

    def __record_completed(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        if value.source() is self:
            self.__datalog.append(ChainedTaskLogEntry(
                value.task().api_id(),
                value.task().uid(),
                ChainedTaskState.finalized
            ))

    def scheduler_feedback(self, scheduler: 'SchedulerProto', feedback: SchedulerFeedback) -> None:
        """ Register a scheduler by this callback
        """

        if feedback == SchedulerFeedback.source_subscribed:
            if self.__scheduler:
                raise ValueError('Unable to subscribe a second scheduler')
            self.__scheduler = scheduler
            self.__scheduler.callback(SchedulerProto.scheduled_task_completed, self.__record_completed_clbk)
        elif feedback == SchedulerFeedback.source_unsubscribed:
            if not self.__scheduler:
                raise ValueError('Unable to unsubscribe unknown scheduler')
            self.__scheduler.remove_callback(SchedulerProto.scheduled_task_completed, self.__record_completed_clbk)
            self.__scheduler = None
        else:
            raise ValueError('Unknown feedback spotted')

    def datalog(self) -> DatalogProto:
        return self.__datalog

    @classmethod
    def wait_for(cls, datalog: DatalogProto, api_id: str) -> typing.Optional[TaskResult]:

        def result_fn(entry: ChainedTaskLogEntry) -> bool:
            return entry.api_id == api_id and entry.state == ChainedTaskState.completed

        def complete_fn(entry: ChainedTaskLogEntry) -> bool:
            return entry.api_id == api_id and entry.state in (ChainedTaskState.completed, ChainedTaskState.finalized)

        result = datalog.find(result_fn, reverse=True)

        if result is None:
            with suppress(StopIteration):
                with SignalWaiter(datalog, DatalogProto.new_entry, value_matcher=complete_fn):
                    if datalog.find(complete_fn, reverse=True):
                        raise StopIteration('Stop to wait')

            result = datalog.find(result_fn, reverse=True)
        return result.result if result is not None else None

    def __record_group_id(self, api_id: str) -> str:
        """ Return identifier that will describe a group of tasks (in order to prevent a parallel run)
        """
        return f'{self.__source_uid}--{api_id}'

    def __skip_started(self, api_ids: typing.Set[str]) -> typing.Set[str]:
        """ Filter a list of api id and return only those that is not running at the moment
        """
        started = set()

        for log_entry in self.__datalog.iterate():
            if log_entry.api_id in api_ids:
                started.add(log_entry.api_id)

        return api_ids.difference(started)

    def __execution_row(self, api_id: str) -> None:
        """ Execute a task with api id and it's dependencies
        """
        assert(self.__queue_proxy.is_inside())

        if self.started_task(api_id) is not None:
            raise ValueError(f'The task "{api_id}" has been started already')

        task_cls = self.__registry.get(api_id)
        unprocessed_deps = [task_cls.dependencies()]
        execution_row = [api_id]

        while unprocessed_deps:
            next_deps = set()

            for dep in unprocessed_deps:
                if dep is None:
                    continue

                required = self.__skip_started(dep)

                if required.intersection(execution_row):
                    raise ValueError('Mutual dependencies found for a task')

                next_deps.update(required)
                for r in required:
                    execution_row.insert(0, r)

            unprocessed_deps = []
            for d in next_deps:
                task_cls = self.__registry.get(d)
                unprocessed_deps.append(task_cls.dependencies())

        for i in execution_row:
            self.__exec(i)

    def __wait_response(self, record: ScheduleRecordProto) -> typing.Optional[ReceivedSignal]:
        """ Start a record and wait for execution result. Result is a signal that has been received for a started
        task. It is one of:
          - SchedulerProto.scheduled_task_started
          - SchedulerProto.scheduled_task_dropped
          - SchedulerProto.scheduled_task_expired

        :param record: a record to start and track
        """
        assert(self.__queue_proxy.is_inside())

        if not self.__scheduler:
            raise ValueError('Scheduler has not registered yet')

        assert(record.postpone_policy() == ScheduledTaskPostponePolicy.drop)

        waiter = SignalWaiter(
            self.__scheduler, SchedulerProto.scheduled_task_started, value_matcher=lambda x: x == record
        )

        self.__scheduler.callback(SchedulerProto.scheduled_task_dropped, waiter)
        self.__scheduler.callback(SchedulerProto.scheduled_task_expired, waiter)

        self.emit(ScheduleSourceProto.task_scheduled, record)
        return waiter.wait()

    def __exec(self, api_id: str) -> None:
        """ Just execute a task with the specified api id
        """
        assert(self.__queue_proxy.is_inside())

        task_cls = self.__registry.get(api_id)
        task_uid = uuid.uuid4()
        record = ScheduleRecord(
            task_cls.create(self.__datalog, api_id, task_uid),
            self,
            simultaneous_runs=1,
            group_id=self.__record_group_id(api_id),
            postpone_policy=ScheduledTaskPostponePolicy.drop
        )

        await_result = self.__wait_response(record)
        if await_result is None or await_result.signal != SchedulerProto.scheduled_task_started:
            raise ValueError('Scheduler refused to start a record')
        else:
            self.__datalog.append(ChainedTaskLogEntry(api_id, task_uid, ChainedTaskState.started))

    def execute(self, api_id: str) -> None:
        """ Request to execute a task (and it's dependencies)
        """
        try:
            self.__queue_proxy.exec(functools.partial(self.__execution_row, api_id), blocking=True)
        except QueueCallbackException as e:
            raise e.__cause__  # type: ignore[misc]  # it's ok

    def started_task(self, api_id: str) -> typing.Optional[ChainedTaskLogEntry]:
        """ Return a last log record of the specified task
        """
        for log_entry in self.__datalog.iterate(reverse=True):
            if log_entry.api_id == api_id:
                return log_entry  # type: ignore[no-any-return]
        return None

    def start(self) -> None:
        """ Start this source
        """
        self.__queue_proxy.start()

    def stop(self) -> None:
        """ Stop this source
        """
        self.__queue_proxy.stop()



# -*- coding: utf-8 -*-
# pyknic/lib/i12n.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code; i12n -- implementation
# TODO: write tests for the code

import typing

from pyknic.lib.registry import APIRegistryProto, register_api


def register_implementation(
    registry: APIRegistryProto, api_id: type
) -> typing.Callable[..., typing.Callable[..., type]]:

    def decorator_fn(decorated_obj: type) -> typing.Callable[..., type]:
        if not issubclass(decorated_obj, api_id):
            raise NotImplementedError(
                f'The "{decorated_obj.__name__}" class must derived from the "{api_id.__name__}" base class'
            )
        return register_api(registry, api_id)(decorated_obj)

    return decorator_fn



# -*- coding: utf-8 -*-
# pyknic/lib/bellboy/models.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

import enum
import typing

import pydantic


@enum.unique
class SecretBackendType(enum.Enum):
    keyring = "keyring"
    shm = "shm"


@enum.unique
class FormattingMode(enum.Enum):
    json = "json"
    rich = "rich"


class CommonBellBoyCommandModel(pydantic.BaseModel):
    formatting: typing.Optional[FormattingMode] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('f', 'formatting'),
        default=None,
        description='the way result will be displayed in console'
    )
    config: typing.Optional[str] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('c', 'config'),
        default=None,
        description='configuration file for this utility'
    )

    model_config = pydantic.ConfigDict(validate_by_alias=True, validate_by_name=True)


class LobbyUrlBellBoyCommandModel(pydantic.BaseModel):

    lobby_url: str = pydantic.Field(
        validation_alias=pydantic.AliasChoices('l', 'lobby-url'),
        description='pyknic server url'
    )

    model_config = pydantic.ConfigDict(validate_by_alias=True, validate_by_name=True)


class SecretBackendBellBoyCommandModel(pydantic.BaseModel):

    secret_backend: SecretBackendType = pydantic.Field(
        validation_alias=pydantic.AliasChoices('s', 'secret-backend'),
        description='defines how temporary session key will be stored '
        '(interaction with the storage is not concurrently safe!)'
    )

    model_config = pydantic.ConfigDict(validate_by_alias=True, validate_by_name=True)


class GeneralBellBoyCommandModel(LobbyUrlBellBoyCommandModel, SecretBackendBellBoyCommandModel):
    model_config = pydantic.ConfigDict(validate_by_alias=True, validate_by_name=True)


class OptionalBellBoyCommandModel(pydantic.BaseModel):
    server: typing.Optional[GeneralBellBoyCommandModel] = None




import os
import re
from setuptools import setup, Extension
from configparser import ConfigParser


__extension_section_re__ = re.compile("^extension=(.*)")

__macro_extension_re__ = re.compile(r"^\s*(\S+)(\s*=\s*(\S+))?")


if __name__ == "__main__":

    ext_modules = []

    config = ConfigParser()
    config.read([os.path.join(os.path.dirname(__file__), 'setup.cfg')])

    for section_name in config.sections():
        r = __extension_section_re__.search(section_name)
        if r is not None:
            section = config[section_name]
            if "sources" not in section:
                raise KeyError('"sources" must be set in "%s" section' % section_name)

            extension_kw = {}

            for option in ["include_dirs", "libraries", "extra_compile_args", "depends"]:
                if option in section:
                    extension_kw[option] = section[option].split()

            if "macros" in section:
                define_macros = []
                for single_record in section["macros"].split("\n"):
                    d = __macro_extension_re__.search(single_record)
                    if d is not None:
                        macro_name, _, macro_value = d.groups()

                        define_macros.append(
                            (macro_name, macro_value if macro_value is not None else "")
                        )
                extension_kw["define_macros"] = define_macros

            e = Extension(r.group(1), section["sources"].split(), **extension_kw)
            ext_modules.append(e)

    setup(ext_modules=(ext_modules if ext_modules else None))



# -*- coding: utf-8 -*-

import typing
import pytest

from inspect import isfunction

from pyknic.lib.verify import Verifier, TypeVerifier, SubclassVerifier, ValueVerifier
from pyknic.lib.verify import verify_value, verify_type, verify_subclass


def sample_function(
    a: typing.Any, b: typing.Any, c: typing.Any, d: typing.Optional[typing.Any] = None, **kwargs: typing.Any
) -> None:
    """ function info """
    pass


class TestVerifier:

    class FNameChecker:

        @staticmethod
        def foo() -> None:
            pass

        @classmethod
        def bar(cls) -> None:
            pass

        def zzz(self) -> None:
            pass

        def __call__(self) -> None:
            pass

    def test_check(self) -> None:
        check = Verifier().check(None, '', lambda x: None)
        assert(isfunction(check) is True)

    def test_decorator(self) -> None:

        verifier = Verifier()

        assert(verifier.decorator()(sample_function) != sample_function)
        verifier = Verifier()

        def exc() -> None:
            raise TypeError('text exception')

        default_check = lambda x: None  # noqa: E731
        exc_check = lambda x: exc() if x == 3 else None  # noqa: E731
        a_spec = lambda x: 111  # noqa: E731
        override_check = lambda s, n, f: exc_check if n == 'a' or n == 'd' or n == 'e' else default_check  # noqa: E731
        verifier.check = override_check  # type: ignore[method-assign] # just an ugly test
        decorated_fn = verifier.decorator(a=a_spec, c=1, d=1, e=1)(sample_function)

        decorated_fn(1, 2, 3, d=4)
        pytest.raises(TypeError, decorated_fn, 3, 2, 3, d=4)
        pytest.raises(TypeError, decorated_fn, 1, 2, 3, d=3)
        decorated_fn(1, 2, 3, d=4, e=5)
        pytest.raises(TypeError, decorated_fn, 1, 2, 3, d=4, e=3)

    def test_function_name(self) -> None:

        assert(Verifier.function_name(TestVerifier.FNameChecker.foo) == 'TestVerifier.FNameChecker.foo')
        assert(Verifier.function_name(TestVerifier.FNameChecker.bar) == 'TestVerifier.FNameChecker.bar')
        assert(Verifier.function_name(TestVerifier.FNameChecker.zzz) == 'TestVerifier.FNameChecker.zzz')
        assert(Verifier.function_name(TestVerifier.FNameChecker) == 'TestVerifier.FNameChecker')

        c = TestVerifier.FNameChecker()
        assert(Verifier.function_name(c.foo) == 'TestVerifier.FNameChecker.foo')
        assert(Verifier.function_name(c.bar) == 'TestVerifier.FNameChecker.bar')
        assert(Verifier.function_name(c.zzz) == 'TestVerifier.FNameChecker.zzz')
        assert(isinstance(Verifier.function_name(c), str))  # we have a result but with random id


class TestTypeVerifier:

    def test_check(self) -> None:
        verifier = TypeVerifier()

        with pytest.raises(RuntimeError):
            verifier.decorator(a=None)(sample_function)

        with pytest.raises(RuntimeError):
            verifier.decorator(a=(str, None, 1))(sample_function)

        decorated_fn = verifier.decorator(
            a=int, b=(str, None), c=[str, int], d=(str, int, None), e=float
        )(sample_function)
        decorated_fn(1, None, 'f')
        decorated_fn(1, None, 1, d='o')
        decorated_fn(1, None, 'o', d=5, e=.1)

        with pytest.raises(TypeError):
            decorated_fn('b', None, 'f')

        with pytest.raises(TypeError):
            decorated_fn(1, 4, 'o')

        with pytest.raises(TypeError):
            decorated_fn(1, None, None)

        with pytest.raises(TypeError):
            decorated_fn(1, None, 'o', d=.1)

        with pytest.raises(TypeError):
            decorated_fn(1, None, 'b', e='a')

        def foo(*args: typing.Any) -> None:
            pass

        decorated_foo = verifier.decorator(args=int)(foo)
        decorated_foo()
        decorated_foo(1, 2, 3)
        pytest.raises(TypeError, decorated_foo, 0.1)
        pytest.raises(TypeError, decorated_foo, 1, 0.1, 4)

        def bar(x: typing.Any, *args: typing.Any) -> None:
            pass

        decorated_bar = verifier.decorator(args=(int, float))(bar)
        decorated_bar('')
        decorated_bar('', 0.1, 2, 3)
        pytest.raises(TypeError, decorated_bar, '', '')


class TestSubclassVerifier:

    def test_check(self) -> None:
        verifier = SubclassVerifier()

        with pytest.raises(RuntimeError):
            verifier.decorator(a=None)(sample_function)

        with pytest.raises(RuntimeError):
            verifier.decorator(a=(str, None, 1))(sample_function)

        class A:
            pass

        class B(A):
            pass

        class C(A):
            pass

        class D:
            pass

        decorated_fn = verifier.decorator(a=A, c=(B, C), d=[B, None], e=C)(sample_function)
        decorated_fn(A, None, B)
        decorated_fn(B, None, B)
        decorated_fn(A, None, B, d=B)
        decorated_fn(A, None, C, d=B)
        decorated_fn(A, None, B, d=None, e=C)

        with pytest.raises(TypeError):
            decorated_fn(D, None, B)

        with pytest.raises(TypeError):
            decorated_fn(A, None, A)

        with pytest.raises(TypeError):
            decorated_fn(A, None, None)

        with pytest.raises(TypeError):
            decorated_fn(A, None, B, d=C)

        with pytest.raises(TypeError):
            decorated_fn(A, None, B, d=B, e=A)


class TestValueVerifier:

    def test_check(self) -> None:

        verifier = ValueVerifier()

        with pytest.raises(RuntimeError):
            verifier.decorator(a=1)(sample_function)

        with pytest.raises(RuntimeError):
            verifier.decorator(a=(1,))(sample_function)

        decorated_fn = verifier.decorator(
            a=(lambda x: x > 5, lambda x: x < 10),
            c=lambda x: x[:3] == 'foo',
            d=lambda x: x is not None,
            e=lambda x: x == 1
        )(sample_function)
        decorated_fn(6, 1, 'foo-asdaads', d=7)
        decorated_fn(6, 1, 'foo-asdaads', d='sss')
        decorated_fn(6, 1, 'foo-asdaads', d=0.1, e=1)

        pytest.raises(ValueError, decorated_fn, 6, None, 'foo-aaa')
        pytest.raises(ValueError, decorated_fn, 5, None, 'foo-aaa', d=5)
        pytest.raises(ValueError, decorated_fn, 10, None, 'foo-aaa', d=5)
        pytest.raises(ValueError, decorated_fn, 6, None, 'foo-aaa', d=5, e=7)


def test_verify() -> None:

    class A:
        a = 'foo'

    @verify_type(a=int, b=str, d=(int, None), e=float)
    @verify_subclass(c=A)
    @verify_value(a=(lambda x: x > 5, lambda x: x < 10), c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
    def foo(
        a: typing.Any, b: typing.Any, c: typing.Any, d: typing.Optional[typing.Any] = None, **kwargs: typing.Any
    ) -> None:
        pass

    foo(6, 'bar', A)
    foo(7, 'bar', A, d=-1)
    foo(6, 'bar', A, e=0.1)

    pytest.raises(TypeError, foo, 'foo', 'bar', A)
    pytest.raises(TypeError, foo, 6, 'bar', int)
    pytest.raises(ValueError, foo, 3, 'bar', A)
    A.a = 'bar'
    pytest.raises(ValueError, foo, 6, 'bar', A)
    A.a = 'foo'
    pytest.raises(ValueError, foo, 6, 'bar', A, d=1)



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/scheduler_executor.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import enum
import functools
import typing

from dataclasses import dataclass
from datetime import datetime, timezone

from pyknic.lib.capability import iscapable
from pyknic.lib.tasks.proto import ScheduleRecordProto, TaskProto, ScheduledTaskPostponePolicy, NoSuchTaskError
from pyknic.lib.tasks.scheduler.queue import SchedulerQueue
from pyknic.lib.signals.proto import SignalSourceProto, Signal
from pyknic.lib.signals.extra import BoundedCallback, SignalResender, CallbacksHolder
from pyknic.lib.signals.proxy import QueueProxy, QueueCallbackException
from pyknic.lib.signals.source import SignalSource
from pyknic.lib.tasks.thread_executor import ThreadExecutor, NoFreeSlotError


class SchedulerExecutor(SignalSource):
    """ This class simplifies SchedulerProto implementation by implementing an execution/postponing process
    as a single point (single source of truth)

    Because it may (and will) be addressed in concurrency manner, it is crucial to follow the steps in order to
    stop a scheduler correctly and in a consistent way. So there they are:
      1. Unsubscribe all the sources in order to stop new record processing.
      The `meth:.SchedulerProto.unsubscribe` method will help
      2. Drop all the postponed tasks. This will do the :meth:`.SchedulerExecutor.cancel_postponed_tasks` method
      3. Request running tasks to stop (the :meth:`.SchedulerExecutor.stop_running_tasks` will do)
      4. Wait for running tasks to complete, join threads and execute all the callbacks that are stored
      (the :meth:`.SchedulerExecutor.await_tasks` method is responsible for this)
    """

    scheduled_task_dropped = Signal(ScheduleRecordProto)    # a scheduled task dropped and would not start
    scheduled_task_postponed = Signal(ScheduleRecordProto)  # a scheduled task postponed and will start later
    scheduled_task_expired = Signal(ScheduleRecordProto)    # a scheduled task dropped because of expired ttl
    scheduled_task_started = Signal(ScheduleRecordProto)    # a scheduled task started
    scheduled_task_completed = Signal(ScheduleRecordProto)  # a scheduled task completed

    @enum.unique
    class TaskState(enum.Enum):
        """ This class shows the state of a task/record
        """
        submitted = enum.auto()  # task is submitted, but decision is not made yet
        pending = enum.auto()    # task is not executed at the moment but waits for it
        started = enum.auto()    # task is started

    @dataclass
    class TaskDescriptor:
        """ Describes a submitted task
        """
        record: ScheduleRecordProto           # record that submitted a task
        state: 'SchedulerExecutor.TaskState'  # state of a task

    def __init__(
        self,
        threads_number: typing.Optional[int] = None,
        executor_cr_timeout: typing.Union[int, float, None] = None,
        thread_cr_timeout: typing.Union[int, float, None] = None
    ):
        """ Create an executor. This also creates an internal queue (:class:`.QueueProxy`) for callbacks processing and
        an internal thread executor (:class:`.ThreadExecutor`) to run tasks

        :param threads_number: same as the "threads_number" argument in :method:`.ThreadExecutor.__init__`
        :param executor_cr_timeout: same as the "executor_cr_timeout" argument in :method:`.ThreadExecutor.__init__`
        :param thread_cr_timeout: same as the "thread_cr_timeout" argument in :method:`.ThreadExecutor.__init__`
        """
        SignalSource.__init__(self)
        self.__proxy = QueueProxy()
        self.__scheduler_queue = SchedulerQueue()
        self.__thread_executor = ThreadExecutor(threads_number, executor_cr_timeout, thread_cr_timeout)

        self.__tasks: typing.Dict[TaskProto, 'SchedulerExecutor.TaskDescriptor'] = dict()

        self.__holder = CallbacksHolder()

        self.__thread_executor.callback(
            ThreadExecutor.task_completed,
            self.__proxy.proxy(
                self.__holder.keep_callback(
                    BoundedCallback(self.__task_completed),
                    self
                )
            )
        )

        self.__scheduler_queue.callback(
            SchedulerQueue.task_expired,
            self.__holder.keep_callback(
                SignalResender(self, target_signal=SchedulerExecutor.scheduled_task_expired),
                self
            )
        )

        self.__scheduler_queue.callback(
            SchedulerQueue.task_dropped,
            self.__holder.keep_callback(
                SignalResender(self, target_signal=SchedulerExecutor.scheduled_task_dropped),
                self
            )
        )

    def __task_completed(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ This callback is executed when a task is about to finish

        :param source: the :class:`.ThreadExecutor` instance
        :param signal: it will be the "ThreadExecutor.task_completed" signal
        :param value: it will be a task (:class:`.TaskProto`) that about to complete
        """
        assert(self.__proxy.is_inside())

        self.__thread_executor.wait_task(value)
        self.__thread_executor.complete_task(value)
        descriptor = self.__tasks.pop(value)
        self.emit(SchedulerExecutor.scheduled_task_completed, descriptor.record)
        self.__proxy.exec(self.__run_postponed_tasks)

    def __run_postponed_tasks(self) -> None:
        """ This callback is for postponed tasks execution
        """
        assert(self.__proxy.is_inside())

        while True:
            try:
                with self.__thread_executor.executor_context() as c:
                    record = self.__scheduler_queue.next_record(self.__filter_record)
                    if record is None:
                        break
                    self.__exec_task(record, c)
            except NoFreeSlotError:
                break

    def __exec_task(self, record: ScheduleRecordProto, context: ThreadExecutor.Context) -> None:
        """ This callback executes a record (a task)

        :param record: a record to execute
        :param context: context with which a slot for execution has been allocated
        """
        assert(self.__proxy.is_inside())

        task = record.task()
        self.__tasks[task].state = SchedulerExecutor.TaskState.started
        context.submit_task(task)
        self.emit(SchedulerExecutor.scheduled_task_started, record)

    def __postpone(self, record: ScheduleRecordProto) -> None:
        """ This callback postpones or drops a record

        :param record: a record to postpone
        """
        assert(self.__proxy.is_inside())

        if record.postpone_policy() == ScheduledTaskPostponePolicy.drop:
            self.__tasks.pop(record.task())
            self.emit(SchedulerExecutor.scheduled_task_dropped, record)
            return

        self.__tasks[record.task()].state = SchedulerExecutor.TaskState.pending
        self.emit(SchedulerExecutor.scheduled_task_postponed, record)
        self.__scheduler_queue.postpone(record)

    def __submit(self, record: ScheduleRecordProto) -> None:
        """ This callback processes a submitted task and tries to execute it

        :param record: a record that should be executed
        """
        assert(self.__proxy.is_inside())

        ttl = record.ttl()
        if ttl is not None and ttl < datetime.now(timezone.utc).timestamp():
            self.emit(SchedulerExecutor.scheduled_task_expired, record)
            return

        if record.task() in self.__tasks:
            raise ValueError('A submitted task is registered already')

        self.__tasks[record.task()] = SchedulerExecutor.TaskDescriptor(
            record, SchedulerExecutor.TaskState.submitted
        )

        if not self.__filter_record(record):
            self.__postpone(record)
            return

        try:
            with self.__thread_executor.executor_context() as c:
                self.__exec_task(record, c)
        except NoFreeSlotError:
            self.__postpone(record)

    def submit(self, record: ScheduleRecordProto, blocking: bool = False) -> None:
        """ Try to execute a record or try to postpone it

        :param record: a record to run
        :param blocking: whether we should wait for result or not (a queue thread may be deadlocked because
        of the True value)
        """
        try:
            self.__proxy.exec(functools.partial(self.__submit, record), blocking=blocking)
        except QueueCallbackException as e:
            raise e.__cause__  # type: ignore[misc]  # it's ok

    def queue_proxy(self) -> QueueProxy:
        """ Return internal proxy
        """
        return self.__proxy

    def __has_tasks(self) -> bool:
        """ Check are there tasks to execute

        :return: True if there are tasks to execute and False otherwise
        """
        return len(self.__scheduler_queue) > 0 or any(self.__thread_executor.tasks())

    def await_tasks(self, task_timeout: typing.Union[int, float, None] = None) -> None:
        """ Wait for tasks to finish

        :param task_timeout: a timeout for each task to complete
        """
        has_tasks = self.__proxy.exec(self.__has_tasks, blocking=True)
        while has_tasks:
            for task in self.__thread_executor.tasks():
                try:
                    if not self.__thread_executor.wait_task(task, timeout=task_timeout):
                        raise TimeoutError('Unable to wait for a task in a time')
                except NoSuchTaskError:
                    # there may be a slight race condition between thread_executor.tasks() and wait_task calls
                    pass

            self.__proxy.exec(self.__run_postponed_tasks, blocking=True)
            has_tasks = self.__proxy.exec(self.__has_tasks, blocking=True)

    def __filter_record(self, record: ScheduleRecordProto) -> bool:
        """ A filter for the :meth:`.SchedulerQueue.next_record` method that checks the "simultaneous_runs" option
        """
        group_id = record.group_id()
        simultaneous_runs = record.simultaneous_runs()

        def sim_runs(x: SchedulerExecutor.TaskDescriptor) -> int:
            return 1 if (x.state == SchedulerExecutor.TaskState.started and x.record.group_id() == group_id) else 0

        if group_id is not None and simultaneous_runs > 0:
            group_count = sum(map(sim_runs, self.__tasks.values()))
            if group_count >= simultaneous_runs:
                return False

        return True

    def __cancel_postponed_tasks(self) -> None:
        """ Cancel all the pending tasks
        """
        assert (self.__proxy.is_inside())

        next_record = self.__scheduler_queue.next_record()
        while next_record:
            self.__tasks.pop(next_record.task())
            self.emit(SchedulerExecutor.scheduled_task_dropped, next_record)
            next_record = self.__scheduler_queue.next_record()

    def cancel_postponed_tasks(self) -> None:
        """ Request to cancel all the pending tasks
        """
        self.__proxy.exec(self.__cancel_postponed_tasks, blocking=True)

    def __stop_running_tasks(self) -> None:
        """ Cancel all the pending tasks
        """
        assert(self.__proxy.is_inside())

        for task in self.__tasks:
            if iscapable(task, TaskProto.stop):
                task.stop()
            elif iscapable(task, TaskProto.terminate):
                task.terminate()

    def stop_running_tasks(self) -> None:
        """ Request to cancel all the running tasks
        """
        self.__proxy.exec(self.__stop_running_tasks, blocking=True)

    def __tasks_filter(
        self,
        filter_fn: typing.Callable[['SchedulerExecutor.TaskDescriptor'], bool]
    ) -> typing.Tuple[TaskProto, ...]:
        """ This method returns tasks which descriptors is suitable

        :param filter_fn: filter function that checks descriptors, this function must return True
        for every suitable task
        """
        assert(self.__proxy.is_inside())

        return tuple((x for x, y in self.__tasks.items() if filter_fn(y)))

    def running_tasks(self) -> typing.Tuple[TaskProto, ...]:
        """ Return tasks that are running at the moment
        """
        return self.__proxy.exec(  # type: ignore[no-any-return]
            functools.partial(
                self.__tasks_filter,
                lambda x: x.state == SchedulerExecutor.TaskState.started
            ),
            blocking=True
        )

    def pending_tasks(self) -> typing.Tuple[TaskProto, ...]:
        """ Return tasks that are waiting for execution
        """
        return self.__proxy.exec(  # type: ignore[no-any-return]
            functools.partial(
                self.__tasks_filter,
                lambda x: x.state != SchedulerExecutor.TaskState.started
            ),
            blocking=True
        )



# -*- coding: utf-8 -*-
# pyknic/lib/io/crypto/cipher.py
#
# Copyright (C) 2016-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import base64
import typing

from abc import abstractmethod

import pydantic

from cryptography.hazmat.primitives.ciphers import Cipher, modes
from cryptography.hazmat.primitives.ciphers.algorithms import AES as C10yAES
from cryptography.hazmat.backends import default_backend

from pyknic.lib.io import IOGenerator, IOProducer
from pyknic.lib.io.aligner import Aligner
from pyknic.lib.registry import APIRegistry, register_api
from pyknic.lib.crypto.random import random_bytes
from pyknic.lib.crypto.proto import CipherProto

# TODO: review modern security recommendations periodically --
#  https://cryptography.io/en/latest/hazmat/primitives/symmetric-encryption/

__default_cipher_registry__ = APIRegistry()


class CipherModeModel(pydantic.BaseModel):
    """This base model helps to describe data that requires for cipher mode initialization.
    """

    @classmethod
    def generate(cls, cipher: CipherProto) -> 'CipherModeModel':
        """Create a cipher mode initialized with random data

        :param cipher: Cipher which will be used with this mode
        """
        raise NotImplementedError('This method is abstract')

    @classmethod
    def deserialize(cls, model_data: typing.Dict[str, typing.Any], cipher: CipherProto) -> 'CipherModeModel':
        """Import dumped models (python dictionary) into this model

        :param model_data: dumped model data
        :param cipher: cipher which will be used with this mode
        """
        raise NotImplementedError('This method is abstract')

    def c10y_mode(self) -> modes.Mode:
        """Create object suitable to use with the "cryptography" package
        """
        raise NotImplementedError('This method is abstract')


class CBCMode(CipherModeModel):
    """The :class:`.CipherModeModel` model implementation that describes CBC-mode

    see also: https://cryptography.io/en/latest/hazmat/primitives/symmetric-encryption/#cryptography.hazmat.primitives.ciphers.modes.CBC  # noqa: E501
    """

    initialization_vector: pydantic.Base64Bytes
    mode: typing.Literal['CBC'] = pydantic.Field(default='CBC', frozen=True)

    @classmethod
    def generate(cls, cipher: CipherProto) -> 'CBCMode':
        """The :meth:`.CipherModeModel.generate` method implementation
        """
        block_size = cipher.algo_block_size()
        assert(block_size is not None)
        return CBCMode(initialization_vector=base64.b64encode(random_bytes(block_size)))

    @classmethod
    def deserialize(cls, model_data: typing.Dict[str, typing.Any], cipher: CipherProto) -> 'CBCMode':
        """The :meth:`.CipherModeModel.deserialize` method implementation
        """
        mode = CBCMode.model_validate(model_data)
        if len(mode.initialization_vector) != cipher.algo_block_size():
            raise ValueError("Invalid initialization vector length")
        return mode

    def c10y_mode(self) -> modes.Mode:
        """The :meth:`.CipherModeModel.c10y_mode` method implementation
        """
        return modes.CBC(self.initialization_vector)


class CTRMode(CipherModeModel):
    """The :class:`.CipherModeModel` model implementation that describes CBC-mode

    see also: https://cryptography.io/en/latest/hazmat/primitives/symmetric-encryption/#cryptography.hazmat.primitives.ciphers.modes.CTR  # noqa: E501

    :note: please note minimal block size!
    """

    nonce: pydantic.Base64Bytes = pydantic.Field(min_length=22)  # (((128 / 8) / 3) * 4) -- length of base64-encoded
    # string required for 128-bit block
    mode: typing.Literal['CTR'] = pydantic.Field(default='CTR', frozen=True)

    @classmethod
    def generate(cls, cipher: CipherProto) -> 'CTRMode':
        """The :meth:`.CipherModeModel.generate` method implementation
        """
        block_size = cipher.algo_block_size()
        assert(block_size is not None)
        return CTRMode(nonce=base64.b64encode(random_bytes(block_size)))

    @classmethod
    def deserialize(cls, model_data: typing.Dict[str, typing.Any], cipher: CipherProto) -> 'CTRMode':
        """The :meth:`.CipherModeModel.deserialize` method implementation
        """
        mode = CTRMode.model_validate(model_data)
        if len(mode.nonce) != cipher.algo_block_size():
            raise ValueError("Invalid nonce size")

        return mode

    def c10y_mode(self) -> modes.Mode:
        """The :meth:`.CipherModeModel.c10y_mode` method implementation
        """
        return modes.CTR(self.nonce)


class _PlainAESCipher(CipherProto):
    """Base class for AES implementation (without key size and without cipher mode)
    """

    @classmethod
    @abstractmethod
    def cipher_mode_model_cls(cls) -> typing.Type[CipherModeModel]:
        """Return model class for cipher mode
        """
        raise NotImplementedError('This method is abstract')

    @classmethod
    def create_encryptor(cls, key: bytes) -> 'CipherProto':
        """The :meth:`.CipherProto.create_encryptor` method implementation
        """
        return cls(key)

    @classmethod
    def create_decryptor(cls, key: bytes, decryptor_init_data: typing.Any) -> 'CipherProto':
        """The :meth:`.CipherProto.create_decryptor` method implementation
        """
        return cls(key, decryptor_init_data)

    @classmethod
    def algo_block_size(cls) -> int:
        """The :meth:`.CipherProto.algo_block_size` method implementation
        """
        return int(128 / 8)

    def __init__(self, key: bytes, init_data: typing.Any = None):
        """Create new encryptor/decryptor

        :param key: key to use
        :param init_data: decryptor's initialization data. If it is None, then this object allows encryption and
        allows decryption otherwise
        """
        CipherProto.__init__(self)

        if len(key) != self.key_size():
            raise ValueError(f"Invalid key size. It must be {self.key_size()} bytes")

        self.__key = key

        m_cls = self.cipher_mode_model_cls()
        self.__mode_model = m_cls.deserialize(init_data, self) if init_data is not None else m_cls.generate(self)
        self.__encryption_mode = init_data is None

        self.__aes_algorithm = C10yAES(key)
        self.__aes_cipher = Cipher(
            self.__aes_algorithm,
            self.__mode_model.c10y_mode(),
            backend=default_backend()
        )

    def decryptor_init_data(self) -> typing.Any:
        """The :meth:`.CipherProto.decryptor_init_data` method implementation
        """
        return self.__mode_model.model_dump()

    def encrypt(self, data: IOProducer) -> IOGenerator:
        """The :meth:`.CipherProto.encrypt` method implementation
        """
        if not self.__encryption_mode:
            raise RuntimeError('Unable to encrypt with decryptor')

        encryptor = self.__aes_cipher.encryptor()  # type: ignore[misc]  # mypy issue
        aligner = Aligner(self.algo_block_size(), strict_mode=True)

        for i in aligner.iterate_data(data):
            yield encryptor.update(i)

        yield encryptor.finalize()

    def decrypt(self, data: IOProducer) -> IOGenerator:
        """The :meth:`.CipherProto.decrypt` method implementation
        """

        if self.__encryption_mode:
            raise RuntimeError('Unable to decrypt with encryptor')

        decryptor = self.__aes_cipher.decryptor()  # type: ignore[misc]  # mypy issue
        aligner = Aligner(self.algo_block_size(), strict_mode=True)

        for i in aligner.iterate_data(data):
            yield decryptor.update(i)

        yield decryptor.finalize()


def _aes_ciphers_generator(
    key_size: int, model_cls: typing.Type[CipherModeModel]
) -> typing.Callable[[typing.Type[_PlainAESCipher]], typing.Type[_PlainAESCipher]]:
    """This function helps to describe specific AES ciphers

    :param key_size: The size of the key to use
    :param model_cls: The cipher mode to use
    """

    def class_decorator(origin_cls: typing.Type[_PlainAESCipher]) -> typing.Type[_PlainAESCipher]:

        class DecoratedClass(origin_cls):  # type: ignore[valid-type, misc]

            @classmethod
            def key_size(cls) -> int:
                return key_size

            @classmethod
            def cipher_mode_model_cls(cls) -> typing.Type[CipherModeModel]:
                return model_cls

        return DecoratedClass

    return class_decorator


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-128-CBC')
@_aes_ciphers_generator(int(128 / 8), CBCMode)
class AES128CBCCipher(_PlainAESCipher):
    """The AES-128-CBC cipher implementation
    """
    pass


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-192-CBC')
@_aes_ciphers_generator(int(192 / 8), CBCMode)
class AES192CBCCipher(_PlainAESCipher):
    """The AES-192-CBC cipher implementation
    """
    pass


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-256-CBC')
@_aes_ciphers_generator(int(256 / 8), CBCMode)
class AES256CBCCipher(_PlainAESCipher):
    """The AES-256-CBC cipher implementation
    """
    pass


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-128-CTR')
@_aes_ciphers_generator(int(128 / 8), CTRMode)
class AES128CTRCipher(_PlainAESCipher):
    """The AES-128-CTR cipher implementation
    """
    pass


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-192-CTR')
@_aes_ciphers_generator(int(192 / 8), CTRMode)
class AES192CTRCipher(_PlainAESCipher):
    """The AES-192-CTR cipher implementation
    """
    pass


# noinspection PyAbstractClass
@register_api(__default_cipher_registry__, 'AES-256-CTR')
@_aes_ciphers_generator(int(256 / 8), CTRMode)
class AES256CTRCipher(_PlainAESCipher):
    """The AES-256-CTR cipher implementation
    """
    pass



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/thread_executor.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import threading
import types
import typing

from pyknic.lib.signals.extra import SignalResender
from pyknic.lib.signals.proto import Signal
from pyknic.lib.signals.source import SignalSource
from pyknic.lib.thread import CriticalResource
from pyknic.lib.verify import verify_value

from pyknic.lib.tasks.proto import TaskExecutorProto, TaskProto, NoSuchTaskError, TaskStartError
from pyknic.lib.tasks.threaded_task import ThreadedTask


class NoFreeSlotError(Exception):
    """ This exception is raised when there is no free slots available
    """
    pass


class ThreadExecutor(TaskExecutorProto, CriticalResource, SignalSource):
    """ The :class:`.TaskExecutorProto` class implementation, that executes tasks in dedicated threads
    """
    # TODO: do not start thread every time but reuse them

    task_completed = Signal(TaskProto)  # this signal is emitted when a task is about to complete

    class Context:
        """ A context that is allocated with a free slot in a pool
        """

        def __init__(
            self,
            executor: 'ThreadExecutor',
            submit_fn: typing.Callable[[TaskProto, bool], ThreadedTask],
            release_fn: typing.Callable[[], None]
        ):
            """ Prepare a context

            :param executor: executor that creates this context
            :param submit_fn: a function that submits a task
            :param release_fn: a function that release a slot if a task wasn't started
            """
            self.__executor = executor
            self.__submit_fn = submit_fn
            self.__release_fn = release_fn
            self.__task_submitted = False

        def __enter__(self) -> 'ThreadExecutor.Context':
            """ Enter a context
            """
            return self

        def __exit__(
            self,
            exc_type: typing.Optional[typing.Type[BaseException]],
            exc_val: typing.Optional[BaseException],
            exc_tb: typing.Optional[types.TracebackType]
        ) -> None:
            """ Exit this context and free slot if a task wasn't started
            """
            if not self.__task_submitted:
                self.__release_fn()

        def submit_task(self, task: TaskProto, start_task: bool = True) -> ThreadedTask:
            """ Try to submit a task

            :param task: a task to start
            :param start_task: whether to start a task immediately
            """
            if self.__task_submitted:
                raise TaskStartError('This context is used already')

            result = self.__submit_fn(task, start_task)
            self.__task_submitted = True
            return result

    @verify_value(threads_number=lambda x: x is None or x > 0)
    def __init__(
        self,
        threads_number: typing.Optional[int] = None,
        executor_cr_timeout: typing.Union[int, float, None] = None,
        thread_cr_timeout: typing.Union[int, float, None] = None
    ):
        """ Create a new executor

        :param threads_number: if defined then this is a maximum number of concurrent running tasks
        :param executor_cr_timeout: this timeout for accessing internal critical resources
        :param thread_cr_timeout: a timeout with which an :class:`.ThreadedTask` object will be created (the same
        as the cr_timeout in the ::meth:`.ThreadedTask.__init__` method)
        """
        TaskExecutorProto.__init__(self)
        CriticalResource.__init__(self, executor_cr_timeout)
        SignalSource.__init__(self)

        self.__threads_semaphore = threading.Semaphore(threads_number) if threads_number else None
        self.__thread_cr_timeout = thread_cr_timeout

        self.__running_threads: typing.Dict[TaskProto, ThreadedTask] = dict()

        self.__signal_resender = SignalResender(self, target_signal=ThreadExecutor.task_completed)

    def __submit_task(self, task: TaskProto, start_task: bool = True) -> ThreadedTask:
        """ Try to start a task

        :param task: a task to start
        """
        task_id = task.task_name() if task.task_name() else str(task)
        result = ThreadedTask(task, self.__thread_cr_timeout, thread_name=f'pyknic:thexec:{task_id}')  # noqa: E231

        with self.critical_context():
            if task in self.__running_threads:
                raise TaskStartError('A task has been submitted already')

            result.callback(ThreadedTask.thread_ready, self.__signal_resender)
            if start_task:
                result.start()

            self.__running_threads[task] = result

        return result

    def __release_slot(self) -> None:
        """ Release a slot (if they are limited)
        """
        if self.__threads_semaphore:
            self.__threads_semaphore.release()

    def submit_task(self, task: TaskProto) -> bool:
        """ The :meth:`.TaskExecutorProto.submit_task` implementation
        """

        try:
            with self.executor_context() as c:
                c.submit_task(task)
                return True
        except (TaskStartError, NoFreeSlotError):
            return False

    def complete_task(self, task: TaskProto) -> bool:
        """ The :meth:`.TaskExecutorProto.complete_task` implementation
        """
        with self.critical_context():
            if task not in self.__running_threads:
                raise NoSuchTaskError('Unable to find a task')
            threaded_task = self.__running_threads[task]

            join_result = threaded_task.join()
            if not join_result:
                return False

            self.__running_threads.pop(task)

        self.__release_slot()
        return True

    def executor_context(self) -> 'ThreadExecutor.Context':
        """ Allocate a slot for a task and return a context
        """
        if self.__threads_semaphore and not self.__threads_semaphore.acquire(False):
            raise NoFreeSlotError("Unable to allocate a slot of the executor's pool")

        return ThreadExecutor.Context(self, self.__submit_task, self.__release_slot)

    def tasks(self) -> typing.Generator[TaskProto, None, None]:
        """ Return started tasks
        """
        with self.critical_context():
            tasks = list(self.__running_threads.keys())

        for i in tasks:
            yield i

    def wait_task(self, task: TaskProto, timeout: typing.Union[int, float, None] = None) -> bool:
        """ The :meth:`.TaskExecutorProto.wait_task` implementation
        """
        with self.critical_context():
            if task not in self.__running_threads:
                raise NoSuchTaskError('Unable to find a task')
            threaded_task = self.__running_threads[task]

        if timeout is None:
            threaded_task.wait()
            return True
        elif timeout > 0:
            threaded_task.wait(timeout=timeout)

        return threaded_task.join()  # type: ignore[no-any-return]  # mypy and decorator's issue

    async def start_async(self, task: TaskProto) -> None:
        """ The :meth:`.TaskExecutorProto.start_async` implementation
        """

        with self.executor_context() as c:
            threaded_task = c.submit_task(task, start_task=False)
            await threaded_task.start_async()



# -*- coding: utf-8 -*-

import pytest

from pyknic.lib.datalog.proto import DatalogProto


def test_abstract() -> None:
    pytest.raises(TypeError, DatalogProto)
    pytest.raises(NotImplementedError, DatalogProto.append, None, None)
    pytest.raises(NotImplementedError, DatalogProto.iterate, None)
    pytest.raises(NotImplementedError, DatalogProto.truncate, None, 0)
    pytest.raises(NotImplementedError, DatalogProto.find, None, 0)



# -*- coding: utf-8 -*-

import asyncio
import pytest
import time
import threading
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks

from pyknic.lib.tasks.thread_executor import ThreadExecutor, NoFreeSlotError
from pyknic.lib.tasks.proto import TaskExecutorProto, NoSuchTaskError, TaskStartError

from fixtures.asyncio import pyknic_async_test


def test_exceptions() -> None:
    assert(issubclass(NoFreeSlotError, Exception) is True)


class TestThreadExecutor:

    def test(self) -> None:
        executor = ThreadExecutor()
        assert(isinstance(executor, TaskExecutorProto) is True)

        assert(set(executor.tasks()) == set())

    @pyknic_async_test
    async def test_exceptions(self, sample_tasks: 'SampleTasks', module_event_loop: asyncio.AbstractEventLoop) -> None:
        executor = ThreadExecutor()

        with pytest.raises(NoSuchTaskError):
            executor.complete_task(sample_tasks.LongRunningTask(terminate_method=False))

        with pytest.raises(NoSuchTaskError):
            executor.wait_task(sample_tasks.LongRunningTask(terminate_method=False))

    def test_join(self, sample_tasks: 'SampleTasks') -> None:
        task = sample_tasks.LongRunningTask(terminate_method=False)
        executor = ThreadExecutor()
        assert(executor.submit_task(task) is True)
        task.stop()

        while not executor.complete_task(task):
            time.sleep(0.1)

    @pytest.mark.filterwarnings("ignore")  # the PytestUnhandledThreadExceptionWarning is a part of the test
    def test_awaited_join(self, sample_tasks: 'SampleTasks') -> None:
        task = sample_tasks.DummyTask()
        executor = ThreadExecutor()
        assert(executor.submit_task(task) is True)

        def thread_fn() -> None:
            nonlocal task  # noqa: F824
            time.sleep(0.5)
            task.stop()

        thread = threading.Thread(target=thread_fn)
        thread.start()

        executor.wait_task(task)
        executor.complete_task(task)

        thread.join()

    def test_awaited_w_timeout_join(self, sample_tasks: 'SampleTasks') -> None:
        task = sample_tasks.LongRunningTask(terminate_method=False)
        executor = ThreadExecutor()
        assert (executor.submit_task(task) is True)

        def thread_fn() -> None:
            nonlocal task  # noqa: F824
            time.sleep(0.5)
            task.stop()

        thread = threading.Thread(target=thread_fn)
        thread.start()

        executor.wait_task(task, timeout=100)
        executor.complete_task(task)

        thread.join()

    @pyknic_async_test
    async def test_async_awaited_join(
        self,
        sample_tasks: 'SampleTasks',
        module_event_loop: asyncio.AbstractEventLoop
    ) -> None:
        task = sample_tasks.LongRunningTask(terminate_method=False)
        executor = ThreadExecutor()

        def thread_fn() -> None:
            nonlocal task  # noqa: F824
            time.sleep(0.5)
            task.stop()

        thread = threading.Thread(target=thread_fn)
        thread.start()

        await executor.start_async(task)
        executor.complete_task(task)

        thread.join()

    def test_threads_number(self, sample_tasks: 'SampleTasks') -> None:
        task1 = sample_tasks.LongRunningTask(terminate_method=False)
        task2 = sample_tasks.LongRunningTask(terminate_method=False)
        task3 = sample_tasks.LongRunningTask(terminate_method=False)

        executor = ThreadExecutor(2)
        assert(executor.submit_task(task1) is True)
        assert(set(executor.tasks()) == {task1})

        assert(executor.submit_task(task1) is False)

        assert(executor.submit_task(task2) is True)
        assert(set(executor.tasks()) == {task1, task2})

        assert(executor.submit_task(task3) is False)

        assert(set(executor.tasks()) == {task1, task2})

        task1.stop()
        task2.stop()

        executor.wait_task(task1)
        executor.wait_task(task2)
        executor.complete_task(task1)
        executor.complete_task(task2)

        assert(set(executor.tasks()) == set())

    def test_context(self, sample_tasks: 'SampleTasks') -> None:
        task1 = sample_tasks.LongRunningTask(terminate_method=False)
        task2 = sample_tasks.LongRunningTask(terminate_method=False)

        executor = ThreadExecutor(2)
        with executor.executor_context():
            with executor.executor_context():
                with pytest.raises(NoFreeSlotError):
                    with executor.executor_context():
                        pass

        with executor.executor_context() as c:
            c.submit_task(task1)
            assert(set(executor.tasks()) == {task1})

            with pytest.raises(TaskStartError):
                c.submit_task(task2)

        task1.stop()
        executor.wait_task(task1)
        executor.complete_task(task1)



# -*- coding: utf-8 -*-
# pyknic/lib/signals/source.py
#
# Copyright (C) 2019-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import threading
import typing

from abc import ABCMeta
from weakref import WeakSet
from inspect import isclass

from pyknic.lib.signals.proto import Signal, SignalSourceProto, UnknownSignalException, SignalCallbackType


class SignalSourceMeta(ABCMeta):
    """ This class helps to manage signals defined for the class
    """

    def __new__(
        mcs,
        name: str,
        bases: typing.Tuple[type],
        namespace: typing.Dict[str, typing.Any], **kwargs: typing.Any
    ) -> type:
        """ Generate new class with this metaclass
        """
        obj = ABCMeta.__new__(mcs, name, bases, namespace, **kwargs)
        obj.__pyknic_signals__ = set()  # type: ignore[attr-defined] # metaclass and mypy issues
        return obj

    def __init__(cls, name: str, bases: typing.Tuple[type], namespace: typing.Dict[str, typing.Any]):
        """ Initialize class with this metaclass
        """
        ABCMeta.__init__(cls, name, bases, namespace)

        for class_attr in dir(cls):
            class_attr_value = ABCMeta.__getattribute__(cls, class_attr)
            if isinstance(class_attr_value, Signal) is True:
                for base_class in bases:
                    try:
                        base_class_attr_value = ABCMeta.__getattribute__(base_class, class_attr)
                        if class_attr_value != base_class_attr_value:
                            raise TypeError(
                                'Signals may not be overridden! Duplicated signal (%s) spotted for the class "%s"'
                                ' (found at the base class %s)'
                                % (class_attr, str(cls), str(base_class))
                            )
                        base_class_attr_value.__pyknic_signal_name__ = f'{base_class.__name__}.{class_attr}'
                    except AttributeError:
                        pass
                if not class_attr_value.__pyknic_signal_name__:
                    class_attr_value.__pyknic_signal_name__ = f'{cls.__name__}.{class_attr}'
                cls.__pyknic_signals__.add(class_attr_value)  # type: ignore[attr-defined] # metaclass and mypy issues


class SignalSource(SignalSourceProto, metaclass=SignalSourceMeta):
    """ :class:`.SignalSourceProto` implementation
    """

    def __init__(self) -> None:
        """ Create new signal source
        """
        SignalSourceProto.__init__(self)

        self.__lock = threading.Lock()

        self.__callbacks: typing.Dict[Signal, WeakSet[SignalCallbackType]] = {
            x: WeakSet() for x in self.__class__.__pyknic_signals__  # type: ignore[attr-defined] # metaclass and mypy
        }

    def emit(self, signal: Signal, signal_value: typing.Any = None) -> None:
        """ :meth:`.SignalSourceProto.emit` implementation
        """
        try:
            with self.__lock:
                callbacks = self.__callbacks[signal].copy()
        except KeyError:
            raise UnknownSignalException('Unknown signal emitted')

        signal.check_value(signal_value)

        for c in callbacks:
            if c is not None:
                c(self, signal, signal_value)

    def callback(self, signal: Signal, callback: SignalCallbackType) -> None:
        """ :meth:`.SignalSourceProto.callback` implementation
        """
        try:
            if hasattr(callback, '__self__') and not isclass(callback.__self__):
                raise ValueError('Bounded methods are unsupported')  # since they are discarded by gc

            with self.__lock:
                self.__callbacks[signal].add(callback)
        except KeyError:
            valid_signals = ', '.join([str(x) for x in self.__callbacks.keys()])
            cls_name = self.__class__.__name__
            raise UnknownSignalException(
                f'Unknown signal subscribed -- {signal} (class -- {cls_name}). Valid signals are: {valid_signals}'
            )

    def remove_callback(self, signal: Signal, callback: SignalCallbackType) -> None:
        """ :meth:`.SignalSourceProto.remove_callback` implementation
        """
        try:
            callbacks = self.__callbacks[signal]
            with self.__lock:
                callbacks.remove(callback)
        except KeyError:
            raise UnknownSignalException('Signal does not have the specified callback')
